"""
This module holds sources for CIFAR100 dataset. Only `TFSource` type source are
implemented for now.
"""
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function
import os

import six.moves.cPickle as pickle
import numpy as np
import tensorflow as tf

from ..core.sources import ClassificationTFSource
from .datasets import DataSet
from six.moves import range


SUPER_CLASS_NUM = 20
SUB_CLASS_NUM = 5
OLD_LABEL_ORDER_LIST = ['apple',
                        'aquarium_fish',
                        'baby',
                        'bear',
                        'beaver',
                        'bed',
                        'bee',
                        'beetle',
                        'bicycle',
                        'bottle',
                        'bowl',
                        'boy',
                        'bridge',
                        'bus',
                        'butterfly',
                        'camel',
                        'can',
                        'castle',
                        'caterpillar',
                        'cattle',
                        'chair',
                        'chimpanzee',
                        'clock',
                        'cloud',
                        'cockroach',
                        'couch',
                        'crab',
                        'crocodile',
                        'cup',
                        'dinosaur',
                        'dolphin',
                        'elephant',
                        'flatfish',
                        'forest',
                        'fox',
                        'girl',
                        'hamster',
                        'house',
                        'kangaroo',
                        'keyboard',
                        'lamp',
                        'lawn_mower',
                        'leopard',
                        'lion',
                        'lizard',
                        'lobster',
                        'man',
                        'maple_tree',
                        'motorcycle',
                        'mountain',
                        'mouse',
                        'mushroom',
                        'oak_tree',
                        'orange',
                        'orchid',
                        'otter',
                        'palm_tree',
                        'pear',
                        'pickup_truck',
                        'pine_tree',
                        'plain',
                        'plate',
                        'poppy',
                        'porcupine',
                        'possum',
                        'rabbit',
                        'raccoon',
                        'ray',
                        'road',
                        'rocket',
                        'rose',
                        'sea',
                        'seal',
                        'shark',
                        'shrew',
                        'skunk',
                        'skyscraper',
                        'snail',
                        'snake',
                        'spider',
                        'squirrel',
                        'streetcar',
                        'sunflower',
                        'sweet_pepper',
                        'table',
                        'tank',
                        'telephone',
                        'television',
                        'tiger',
                        'tractor',
                        'train',
                        'trout',
                        'tulip',
                        'turtle',
                        'wardrobe',
                        'whale',
                        'willow_tree',
                        'wolf',
                        'woman',
                        'worm']

# Five classes in a line is a super class. The order is by the sequences shown
# given in the dataset [website](https://www.cs.toronto.edu/~kriz/cifar.html),
# and is copied in the class comment of `HCifar100TFSource`.
NEW_LABEL_ORDER_LIST = [
    'beaver', 'dolphin', 'otter', 'seal', 'whale',
    'aquarium_fish', 'flatfish', 'ray', 'shark', 'trout',
    'orchid', 'poppy', 'rose', 'sunflower', 'tulip',
    'bottle', 'bowl', 'can', 'cup', 'plate',
    'apple', 'mushroom', 'orange', 'pear', 'sweet_pepper',
    'clock', 'keyboard', 'lamp', 'telephone', 'television',
    'bed', 'chair', 'couch', 'table', 'wardrobe',
    'bee', 'beetle', 'butterfly', 'caterpillar', 'cockroach',
    'bear', 'leopard', 'lion', 'tiger', 'wolf',
    'bridge', 'castle', 'house', 'road', 'skyscraper',
    'cloud', 'forest', 'mountain', 'plain', 'sea',
    'camel', 'cattle', 'chimpanzee', 'elephant', 'kangaroo',
    'fox', 'porcupine', 'possum', 'raccoon', 'skunk',
    'crab', 'lobster', 'snail', 'spider', 'worm',
    'baby', 'boy', 'girl', 'man', 'woman',
    'crocodile', 'dinosaur', 'lizard', 'snake', 'turtle',
    'hamster', 'mouse', 'rabbit', 'shrew', 'squirrel',
    'maple_tree', 'oak_tree', 'palm_tree', 'pine_tree', 'willow_tree',
    'bicycle', 'bus', 'motorcycle', 'pickup_truck', 'train',
    'lawn_mower', 'rocket', 'streetcar', 'tank', 'tractor'
]


class Cifar100TFSource(ClassificationTFSource):
    """
    A concrete `Source` for Cifar100 dataset. This class provides global
    contrast normalized, then ZCA whitened images using tfrecords.
    """
    SAMPLE_NUM = 50000

    def _load_cifar100_python(self, filenames):
        """
        Load python version of Cifar100 dataset.
        """
        # Load the first batch of data to get shape info.
        filename = filenames[0]
        with open(filename, "rb") as f:
            tmp = pickle.load(f)
            data = tmp["data"]
            labels = np.array(tmp["fine_labels"])

        # Load the rest.
        for filename in filenames[1:]:
            with open(filename, "rb") as f:
                tmp = pickle.load(f)
                data = np.append(data, tmp["data"], 1)
                labels = np.append(labels, tmp["fine_labels"])

        data = np.reshape(data, [-1, 3, 32, 32])
        data = np.einsum("nchw->nhwc", data)

        return DataSet(data, labels)

    def _forward(self):
        """
        Construct input for CIFAR100 evaluation using the Reader ops.
        """
        self._maybe_convert_to_tf()
        self._read_from_tfrecord()

    def _read_from_tfrecord(self):
        # Read and set up data tensors.
        filename = os.path.join(self.work_dir, 'cifar100_training.tfrecords')
        with tf.name_scope('input'):
            filename_queue = tf.train.string_input_producer([filename])
        self._training_datum, self._training_label \
            = self._get_sample_tensors_from_tfrecords(filename_queue)

        filename = os.path.join(self.work_dir, 'cifar100_test.tfrecords')
        with tf.name_scope('input'):
            filename_queue = tf.train.string_input_producer([filename])
        self._val_datum, self._val_label \
            = self._get_sample_tensors_from_tfrecords(filename_queue)

    def _maybe_convert_to_tf(self):
        """
        If tfrecords data are not available, convert it from the ZCA whitened
        data and downloaded labels.

        TODO: write the data pre-processing code so I won't need to rely on
        pylearn2.
        """
        TRAINING_TF_FILENAME = "cifar100_training"
        if not os.path.exists(
                os.path.join(self.work_dir,
                             TRAINING_TF_FILENAME + ".tfrecords")):
            # Read the numpy data in and convert it to TFRecord.
            imgs = np.load(os.path.join(self.work_dir,
                                        "pylearn2_gcn_whitened",
                                        "train.npy"))
            imgs = imgs.reshape([Cifar100TFSource.SAMPLE_NUM, 3, 32, 32])
            imgs = np.einsum("nchw->nhwc", imgs)

            # Load training set into memory.
            train_filenames = [os.path.join(self.work_dir,
                                            'cifar-100-python',
                                            "train")]
            training_labels = self._load_cifar100_python(
                train_filenames).labels
            self._convert_to_tf(imgs, training_labels, TRAINING_TF_FILENAME)

        TEST_TF_FILENAME = "cifar100_test"
        if not os.path.exists(
                os.path.join(self.work_dir, TEST_TF_FILENAME + ".tfrecords")):
            imgs = np.load(os.path.join(self.work_dir,
                                        "pylearn2_gcn_whitened",
                                        "test.npy"))
            imgs = imgs.reshape([10000, 3, 32, 32])
            imgs = np.einsum("nchw->nhwc", imgs)
            test_filenames = [os.path.join(self.work_dir,
                                           'cifar-100-python',
                                           'test')]
            test_labels = self._load_cifar100_python(test_filenames).labels
            self._convert_to_tf(imgs, test_labels, TEST_TF_FILENAME)

    def _get_sample_tensors_from_tfrecords(self, filename_queue):
        """
        Read from tfrecord file and return data tensors.

        Args:
            filename_queue: tf.train.string_input_producer
                A file name queue that gives string tensor for tfrecord names.

        Returns:
            (image, label): tuple of (rank-4 tf.float32 tensor and rank-1
                            tf.int32 tensor)
                individual sample that may be later put into a batch.
        """
        reader = tf.TFRecordReader()
        _, serialized_example = reader.read(filename_queue)
        features = tf.parse_single_example(
            serialized_example,
            # Defaults are not specified since both keys are required.
            features={
                'image_raw': tf.FixedLenFeature(
                    [32, 32, 3],
                    tf.float32),
                'label': tf.FixedLenFeature([], tf.int64),
            })

        # Convert label from a scalar uint8 tensor to an int32 scalar.
        label = tf.cast(features['label'], tf.int32)
        image = features["image_raw"]

        return image, label


class HCifar100TFSource(ClassificationTFSource):
    """
    A concrete `Source` for Cifar100 dataset to provide both fine labels and
    coarse labels. It remaps fine labels to let classes in a super class has
    consecutive labels. This class provides global contrast normalized, then
    ZCA whitened images using tfrecords.

    Label property of this source will point to a list of three labels, being
    coarse label, fine label and expanded fine label respectively.

    The logic to remap fine labels is as following:

    Alphabetically, coarse labels are assigned to super class (which is
    already done in the original dataset). Starting from super class 0,
    sub-class in this super class is assigned label alphabetically.

    +------------------------------+------------------------------+
    |Superclass                    |Classes                       |
    +------------------------------+------------------------------+
    |aquatic mammals               |beaver, dolphin,  otter, seal,|
    |                              |whale                         |
    +------------------------------+------------------------------+
    |fish                          |aquarium fish,  flatfish, ray,|
    |                              |shark, trout                  |
    +------------------------------+------------------------------+
    |flowers                       |orchids,    poppies,    roses,|
    |                              |sunflowers, tulips            |
    +------------------------------+------------------------------+
    |food containers               |bottles,  bowls,  cans,  cups,|
    |                              |plates                        |
    +------------------------------+------------------------------+
    |fruit and vegetables          |apples,   mushrooms,  oranges,|
    |                              |pears, sweet peppers          |
    +------------------------------+------------------------------+
    |household electrical devices  |clock,    computer   keyboard,|
    |                              |lamp, telephone, television   |
    +------------------------------+------------------------------+
    |household furniture           |bed,   chair,  couch,   table,|
    |                              |wardrobe                      |
    +------------------------------+------------------------------+
    |insects                       |bee,     beetle,    butterfly,|
    |                              |caterpillar, cockroach        |
    +------------------------------+------------------------------+
    |large carnivores              |bear,  leopard,  lion,  tiger,|
    |                              |wolf                          |
    +------------------------------+------------------------------+
    |large man-made outdoor things |bridge,  castle, house,  road,|
    |                              |skyscraper                    |
    +------------------------------+------------------------------+
    |large natural outdoor scenes  |cloud,    forest,    mountain,|
    |                              |plain, sea                    |
    +------------------------------+------------------------------+
    |large omnivores and herbivores|camel,   cattle,   chimpanzee,|
    |                              |elephant, kangaroo            |
    +------------------------------+------------------------------+
    |medium-sized mammals          |fox,     porcupine,    possum,|
    |                              |raccoon, skunk                |
    +------------------------------+------------------------------+
    |non-insect invertebrates      |crab, lobster,  snail, spider,|
    |                              |worm                          |
    +------------------------------+------------------------------+
    |people                        |baby, boy, girl, man, woman   |
    +------------------------------+------------------------------+
    |reptiles                      |crocodile,  dinosaur,  lizard,|
    |                              |snake, turtle                 |
    +------------------------------+------------------------------+
    |small mammals                 |hamster, mouse, rabbit, shrew,|
    |                              |squirrel                      |
    +------------------------------+------------------------------+
    |trees                         |maple, oak, palm, pine, willow|
    +------------------------------+------------------------------+
    |vehicles 1                    |bicycle,    bus,   motorcycle,|
    |                              |pickup truck, train           |
    +------------------------------+------------------------------+
    |vehicles 2                    |lawn-mower, rocket, streetcar,|
    |                              |tank, tractor                 |
    +------------------------------+------------------------------+
    """
    SAMPLE_NUM = 50000
    TRAINING_TF_FILENAME = "hierarchical_cifar100_training.tfrecords"
    TEST_TF_FILENAME = "hierarchical_cifar100_test.tfrecords"

    def __init__(self, **kwargs):
        super(HCifar100TFSource, self).__init__(**kwargs)
        # Construct remapping table.
        remap_list = [-1] * 100
        for i, label in enumerate(OLD_LABEL_ORDER_LIST):
            for j, new_label in enumerate(NEW_LABEL_ORDER_LIST):
                if new_label == label:
                    remap_list[i] = j

        for l in remap_list:
            assert l is not -1, "All label should have a mapped value."

        self.remap_list = remap_list

    def _load_cifar100_python(self, filenames):
        """
        Load python version of Cifar100 dataset.
        """
        # Load the first batch of data to get shape info.
        filename = filenames[0]
        with open(filename, "rb") as f:
            tmp = pickle.load(f)
            fine_labels = np.array(tmp["fine_labels"])
            coarse_labels = np.array(tmp["coarse_labels"])

        # Load the rest.
        for filename in filenames[1:]:
            with open(filename, "rb") as f:
                tmp = pickle.load(f)
                fine_labels = np.append(fine_labels, tmp["fine_labels"])
                coarse_labels = np.append(coarse_labels, tmp["coarse_labels"])

        return fine_labels, coarse_labels

    def _forward(self):
        """
        Construct input for CIFAR100 evaluation using the Reader ops.
        """
        self._maybe_convert_to_tf()
        self._read_from_tfrecord()

    def _read_from_tfrecord(self):
        # Read and set up data tensors.
        filename = os.path.join(self.work_dir,
                                HCifar100TFSource.TRAINING_TF_FILENAME)
        with tf.name_scope('input'):
            filename_queue = tf.train.string_input_producer([filename])
        self._training_datum, self._training_label \
            = self._get_sample_tensors_from_tfrecords(filename_queue)

        filename = os.path.join(self.work_dir,
                                HCifar100TFSource.TEST_TF_FILENAME)
        with tf.name_scope('input'):
            filename_queue = tf.train.string_input_producer([filename])
        self._val_datum, self._val_label \
            = self._get_sample_tensors_from_tfrecords(filename_queue)

    def _maybe_convert_to_tf(self):
        """
        If tfrecords data are not available, convert it from the ZCA whitened
        data and downloaded labels.

        TODO: write the data pre-processing code so I won't need to rely on
        pylearn2.
        """
        if not os.path.exists(
                os.path.join(
                    self.work_dir,
                    HCifar100TFSource.TRAINING_TF_FILENAME)):
            # Read the numpy data in and convert it to TFRecord.
            imgs = np.load(os.path.join(self.work_dir,
                                        "pylearn2_gcn_whitened",
                                        "train.npy"))
            imgs = imgs.reshape([Cifar100TFSource.SAMPLE_NUM, 3, 32, 32])
            imgs = np.einsum("nchw->nhwc", imgs)

            # Load training set into memory.
            train_filenames = [os.path.join(self.work_dir,
                                            'cifar-100-python',
                                            "train")]
            training_fine_labels, training_coarse_labels \
                = self._load_cifar100_python(train_filenames)
            self._remap_fine_labels(training_fine_labels)
            self._convert_to_tf(imgs,
                                training_fine_labels,
                                training_coarse_labels,
                                HCifar100TFSource.TRAINING_TF_FILENAME)

        if not os.path.exists(
                os.path.join(
                    self.work_dir,
                    HCifar100TFSource.TEST_TF_FILENAME)):
            imgs = np.load(os.path.join(self.work_dir,
                                        "pylearn2_gcn_whitened",
                                        "test.npy"))
            imgs = imgs.reshape([10000, 3, 32, 32])
            imgs = np.einsum("nchw->nhwc", imgs)
            test_filenames = [os.path.join(self.work_dir,
                                           'cifar-100-python',
                                           'test')]

            test_fine_labels, test_coarse_labels \
                = self._load_cifar100_python(test_filenames)
            self._remap_fine_labels(test_fine_labels)
            self._convert_to_tf(imgs,
                                test_fine_labels,
                                test_coarse_labels,
                                HCifar100TFSource.TEST_TF_FILENAME)

    def _remap_fine_labels(self, labels):
        """
        Remap labels of sub-class in place.

        Args:
            labels: numpy array of int of shape [N].
        """
        for i, l in enumerate(labels):
            labels[i] = self.remap_list[l]

    def _convert_to_tf(self, images, fine_labels, coarse_labels, name):
        """
        Convert images and labels to tfrecord format.
        """
        num_examples = fine_labels.shape[0]
        if images.shape[0] != num_examples:
            raise ValueError("Images size %d does not match fine label size"
                             "%d." % (images.shape[0], num_examples))
        if coarse_labels.shape[0] != num_examples:
            raise ValueError("Coarse label size %d does not match fine label"
                             " size %d" % (images.shape[0], num_examples))

        expanded_fine_labels = self._expand_fine_labels(fine_labels)

        row = images.shape[1]
        col = images.shape[2]
        depth = images.shape[3]

        filename = os.path.join(self.work_dir, name)
        print(('Writing', filename))
        writer = tf.python_io.TFRecordWriter(filename)
        for index in range(num_examples):
            image_raw = np.reshape(images[index], -1).tolist()
            expanded_fine_label = expanded_fine_labels[index].tolist()
            example = tf.train.Example(features=tf.train.Features(
                feature={
                    'height': self._int_feature([row]),
                    'width': self._int_feature([col]),
                    'depth': self._int_feature([depth]),
                    'coarse_label': self._int_feature(
                        [int(coarse_labels[index])]),
                    'expanded_fine_label': self._int_feature(
                        expanded_fine_label),
                    'fine_label': self._int_feature(
                        [int(fine_labels[index])]),
                    'image_raw': self._float_feature(image_raw)}))
            writer.write(example.SerializeToString())
        writer.close()

    def _expand_fine_labels(self, labels):
        """
        A fine label will be converted to a `(SUB_CLASS_NUM+1) *
        SUPER_CLASS_NUM` vector, where in the super class group the sub-class
        belongs to, the corresponding dimension will be 1, and in other super
        class groups, the non-existence dimension will be 1. In all, in total,
        there will be super-class number of 1 in the label vector whatever.
        """
        num_examples = labels.shape[0]
        fine_label_vectors = np.zeros(
            [num_examples, (SUB_CLASS_NUM+1)*SUPER_CLASS_NUM],
            dtype=np.int)
        # Set default non-existence labels.
        fine_label_vectors[:, SUB_CLASS_NUM::(SUB_CLASS_NUM+1)] = 1
        # Set real labels.
        batch_indices = list(range(0, num_examples))
        augmented_labels = labels // SUB_CLASS_NUM * (SUB_CLASS_NUM+1) \
            + labels % SUB_CLASS_NUM
        fine_label_vectors[batch_indices, augmented_labels] = 1
        # Unset non-existence labels corresponding to real labels.
        labels_to_unset = labels // SUB_CLASS_NUM * (SUB_CLASS_NUM+1) + \
            SUB_CLASS_NUM
        fine_label_vectors[batch_indices, labels_to_unset] = 0

        return fine_label_vectors

    def _get_sample_tensors_from_tfrecords(self, filename_queue):
        """
        Read from tfrecord file and return data tensors.

        Args:
            filename_queue: tf.train.string_input_producer
                A file name queue that gives string tensor for tfrecord names.

        Returns:
            A tuple of a tensor and a list. The tensor is the image, the list
            contains tensors of coarse label, fine label and expanded fine
            labels respectively.
        """
        reader = tf.TFRecordReader()
        _, serialized_example = reader.read(filename_queue)
        features = tf.parse_single_example(
            serialized_example,
            # Defaults are not specified since both keys are required.
            features={
                'image_raw': tf.FixedLenFeature(
                    [32, 32, 3],
                    tf.float32),
                'coarse_label': tf.FixedLenFeature([], tf.int64),
                'fine_label': tf.FixedLenFeature([], tf.int64),
                'expanded_fine_label': tf.FixedLenFeature(
                    [(SUB_CLASS_NUM+1)*SUPER_CLASS_NUM],
                    tf.int64)
            })

        # Convert label from a scalar uint8 tensor to an int32 scalar.
        coarse_label = tf.cast(features['coarse_label'], tf.int32)
        fine_label = tf.cast(features['fine_label'], tf.int32)
        expanded_fine_label = tf.cast(
            features['expanded_fine_label'], tf.int32)
        image = features["image_raw"]

        return image, [coarse_label, fine_label, expanded_fine_label]



from akid import backend as A

@A.cache_name_if_exist
def binary_accuracy(x, y, hinge_loss_label=False, name=None):
    """
    If `hinge_loss_label` is true, the function assumes that elements in the
    label `y` is given as -1, or 1, and 0 is used as the threshold to tell
    apart classes. Otherwise, the label is assume to be given
    in 1, or 0, and 0.5 is used as the threshold.
    """
    x = A.value(x)

    if hinge_loss_label:
        x[x > 0] = 1
        x[x < 0] = -1
    else:
        x[x >= 0.5] = 1
        x[x < 0.5] = 0

    acc = A.mean(A.cast(x == y, A.float32))
    return acc



from __future__ import absolute_import
from akid import AKID_DATA_PATH
from akid.core import kids, kongfus
from akid.datasets import Cifar10FeedSource
from akid.core.sensors import FeedSensor
from akid.core.brains import GraphBrain
from akid.layers import (
    ConvolutionLayer,
    PoolingLayer,
    InnerProductLayer,
    SoftmaxWithLossLayer,
    DropoutLayer,
    CollapseOutLayer
)


def setup(graph, lr):
    # Set up brain
    # #########################################################################
    brain = GraphBrain(name='maxout-relu-cifar10')

    brain.attach(ConvolutionLayer([8, 8],
                                  [1, 1, 1, 1],
                                  'SAME',
                                  init_para={
                                      "name": "uniform",
                                      "range": 0.005},
                                  wd={"type": "l2", "scale": 0.0005},
                                  out_channel_num=192,
                                  name='conv1'))
    brain.attach(PoolingLayer([1, 4, 4, 1],
                              [1, 2, 2, 1],
                              'SAME',
                              name='pool1'))
    brain.attach(CollapseOutLayer(name='maxout1'))
    brain.attach(DropoutLayer(keep_prob=0.8, name='dropout1'))

    brain.attach(ConvolutionLayer([8, 8],
                                  [1, 1, 1, 1],
                                  'SAME',
                                  init_para={
                                      "name": "uniform",
                                      "range": 0.005},
                                  wd={"type": "l2", "scale": 0.0005},
                                  out_channel_num=384,
                                  name='conv2'))
    brain.attach(PoolingLayer([1, 4, 4, 1],
                              [1, 2, 2, 1],
                              'SAME',
                              name='pool2'))
    brain.attach(CollapseOutLayer(name='maxout2'))
    brain.attach(DropoutLayer(keep_prob=0.5, name='dropout2'))

    brain.attach(ConvolutionLayer([5, 5],
                                  [1, 1, 1, 1],
                                  'SAME',
                                  init_para={
                                      "name": "uniform",
                                      "range": 0.005},
                                  wd={"type": "l2", "scale": 0.0005},
                                  out_channel_num=384,
                                  name='conv3'))
    brain.attach(PoolingLayer([1, 2, 2, 1],
                              [1, 2, 2, 1],
                              'SAME',
                              name='pool3'))
    brain.attach(CollapseOutLayer(name='maxout3'))
    brain.attach(DropoutLayer(keep_prob=0.5, name='dropout3'))

    brain.attach(InnerProductLayer(init_para={"name": "uniform",
                                              "range": 0.005},
                                   wd={"type": "l2", "scale": 0.004},
                                   out_channel_num=2500,
                                   name='ip1'))
    brain.attach(CollapseOutLayer(group_size=5, name='maxout4'))
    brain.attach(DropoutLayer(keep_prob=0.3, name='dropout3'))

    brain.attach(InnerProductLayer(init_para={"name": "uniform",
                                              "range": 1/500.0},
                                   wd={"type": "l2", "scale": 0.004},
                                   out_channel_num=10,
                                   name='softmax_linear'))
    brain.attach(SoftmaxWithLossLayer(class_num=10, name='loss'))

    # Set up a sensor.
    # #########################################################################
    cifar_source = Cifar10FeedSource(
        name="CIFAR10",
        url='http://www.cs.toronto.edu/~kriz/cifar-10-binary.tar.gz',
        work_dir=AKID_DATA_PATH + '/cifar10',
        use_zca=True,
        num_train=50000,
        num_val=10000)

    sensor = FeedSensor(source_in=cifar_source,
                        batch_size=128,
                        name='data')

    # Summon a survivor.
    # #########################################################################
    survivor = kids.Kid(
        sensor,
        brain,
        kongfus.MomentumKongFu(base_lr=lr,
                               momentum=0.5,
                               decay_rate=0.1,
                               decay_epoch_num=50),
        graph=graph,
        max_steps=200000)
    survivor.setup()

    return survivor



"""Builds the MNIST network.

This network is a modified network that combine convolutional neural
network(convolutional.py) and MLP(mnist.py) in the tutorial.

Implements the inference/loss/training pattern for model building.

1. inference() - Builds the model as far as is required for running the network
forward to make predictions.
2. loss() - Adds to the inference model the layers required to generate loss.
3. training() - Adds to the loss model the Ops required to generate and
apply gradients.

TensorFlow install instructions:
https://tensorflow.org/get_started/os_setup.html

MNIST tutorial:
https://tensorflow.org/tutorials/mnist/tf/index.html
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math

import tensorflow.python.platform
import tensorflow as tf

# The MNIST dataset has 10 classes, representing the digits 0 through 9.
NUM_CLASSES = 10

# The MNIST images are always 28x28 pixels.
IMAGE_SIZE = 28
IMAGE_PIXELS = IMAGE_SIZE * IMAGE_SIZE
SEED = 66478  # Set to None for random seed.
WEIGHT_DECAY = 5e-4
NUM_TRAIN = 50000
NUM_EVAL = 10000


def _variable_with_weight_decay(name, shape, stddev, wd):
    """Helper to create an initialized Variable with weight decay.

    Note that the Variable is initialized with a truncated normal distribution.
    A weight decay is added only if one is specified.

    Args:
        name: name of the variable
        shape: list of ints
        stddev: standard deviation of a truncated Gaussian
        wd: add L2Loss weight decay multiplied by this float. If None, weight
            decay is not added for this Variable.

    Returns:
        Variable Tensor
    """
    var = tf.Variable(
        tf.truncated_normal(shape, stddev=stddev, seed=SEED),
        name=name)
    if wd:
        weight_decay = tf.multiply(tf.nn.l2_loss(var), wd, name='weight_loss')
        tf.add_to_collection('losses', weight_decay)
    return var


def _activation_summary(x):
    """Helper to create summaries for activations.

    Creates a summary that provides a histogram of activations.
    Creates a summary that measure the sparsity of activations.

    Args:
        x: Tensor
    Returns:
        nothing
    """
    tf.summary.histogram(x.op.name + '/activations', x)
    tf.summary.scalar(x.op.name + '/sparsity', tf.nn.zero_fraction(x))


def inference(images):
    """Build the MNIST model up to where it may be used for inference.

    Args:
        images: Images placeholder, from inputs().

    Returns:
        softmax_linear: Output tensor with the computed logits.
    """
    # Conv 1
    with tf.name_scope('conv1'):
        kernel = _variable_with_weight_decay("weights", [5, 5, 1, 32],
                                             0.1, WEIGHT_DECAY)
        bias = tf.Variable(tf.zeros([32]), name='biases')
        conv = tf.nn.conv2d(images, kernel,
                            strides=[1, 1, 1, 1], padding='SAME',
                            name='conv')
        fmap = tf.nn.bias_add(conv, bias, name='fmap')
        relu = tf.nn.relu(fmap, name='relu')
        pool = tf.nn.max_pool(relu,
                              ksize=[1, 2, 2, 1],
                              strides=[1, 2, 2, 1],
                              padding='SAME',
                              name='pool')
        _activation_summary(pool)
    # Conv 2
    with tf.name_scope('conv2'):
        kernel = _variable_with_weight_decay("weights", [5, 5, 32, 64],
                                             0.1, WEIGHT_DECAY)
        bias = tf.Variable(tf.constant(0.1, shape=[64]), name='biases')
        conv = tf.nn.conv2d(pool, kernel,
                            strides=[1, 1, 1, 1], padding='SAME',
                            name='conv')
        fmap = tf.nn.bias_add(conv, bias, name='fmap')
        relu = tf.nn.relu(fmap, name='relu')
        pool = tf.nn.max_pool(relu,
                              ksize=[1, 2, 2, 1],
                              strides=[1, 2, 2, 1],
                              padding='SAME',
                              name='pool')
        _activation_summary(pool)
    # Reshape the feature map cuboid into a 2D matrix to feed it to the
    # fully connected layers.
    pool_shape = pool.get_shape().as_list()
    flattened_pool = tf.reshape(
        pool,
        [pool_shape[0], pool_shape[1] * pool_shape[2] * pool_shape[3]])
    flattened_pool_shape = flattened_pool.get_shape().as_list()
    # FC1
    with tf.name_scope('fc1'):
        fc = _variable_with_weight_decay('fc1', [flattened_pool_shape[1], 512],
                                         0.1, WEIGHT_DECAY)
        bias = tf.Variable(tf.constant(0.1, shape=[512]), name='biases')
        # Fully connected layer. Note that the '+' operation automatically
        # broadcasts the biases.
        hidden = tf.nn.relu(tf.matmul(flattened_pool, fc) + bias,
                            name='hidden')
        _activation_summary(hidden)
    # FC2
    with tf.name_scope('fc2'):
        fc = _variable_with_weight_decay('fc2', [512, NUM_CLASSES],
                                         0.1, WEIGHT_DECAY)
        bias = tf.Variable(tf.constant(0.1, shape=[NUM_CLASSES]),
                           name='biases')
        logit = tf.nn.relu(tf.matmul(hidden, fc) + bias, name='hidden')
        _activation_summary(logit)
    return logit


def loss(logits, labels):
    """Calculates the loss from the logits and the labels.

    Args:
        logits: Logits tensor, float - [batch_size, NUM_CLASSES].
        labels: Labels tensor, int32 - [batch_size].

    Returns:
        loss: Loss tensor of type float.
    """
    # Convert from sparse integer labels in the range [0, NUM_CLASSSES)
    # to 1-hot dense float vectors (that is we will have batch_size vectors,
    # each with NUM_CLASSES values, all of which are 0.0 except there will
    # be a 1.0 in the entry corresponding to the label).
    batch_size = tf.size(labels)
    labels = tf.expand_dims(labels, 1)
    indices = tf.expand_dims(tf.range(0, batch_size, 1), 1)
    concated = tf.concat(axis=1, values=[indices, labels])
    onehot_labels = tf.sparse_to_dense(
        concated, tf.stack([batch_size, NUM_CLASSES]), 1.0, 0.0)
    cross_entropy = tf.nn.softmax_cross_entropy_with_logits(logits=logits,
                                                            labels=onehot_labels,
                                                            name='xentropy')
    cross_entropy_mean = tf.reduce_mean(cross_entropy, name='xentropy_mean')

    tf.add_to_collection('losses', cross_entropy_mean)

    # The total loss is defined as the cross entropy loss plus all of the
    # weight decay terms (L2 loss).
    loss = tf.add_n(tf.get_collection('losses'), name='total_loss')
    return loss


def training(loss, BATCH_SIZE):
    """Sets up the training Ops.

    Creates a summarizer to track the loss over time in TensorBoard.

    Creates an optimizer and applies the gradients to all trainable variables.

    The Op returned by this function is what must be passed to the
    `sess.run()` call to cause the model to train.

    Args:
        loss: Loss tensor, from loss().
        learning_rate: The learning rate to use for gradient descent.

    Returns:
        train_op: The Op for training.
    """
    # Add a scalar summary for the snapshot loss.
    tf.summary.scalar(loss.op.name, loss)
    # Optimizer: set up a variable that's incremented once per batch and
    # controls the learning rate decay.
    step = tf.Variable(0, name='step', trainable=False)
    # Decay once per epoch, using an exponential schedule starting at 0.01.
    learning_rate = tf.train.exponential_decay(
        0.01,                # Base learning rate.
        step * BATCH_SIZE,  # Current index into the dataset.
        NUM_TRAIN,          # Decay step.
        0.95,                # Decay rate.
        staircase=True)
    tf.summary.scalar('learning_rate', learning_rate)
    # Use simple momentum for the optimization.
    optimizer = tf.train.MomentumOptimizer(learning_rate, 0.9)
    # Create a variable to track the global step.
    # Use the optimizer to apply the gradients that minimize the loss
    # (and also increment the global step counter) as a single training step.
    train_op = optimizer.minimize(loss, global_step=step)
    # Add histograms for trainable variables.
    for var in tf.trainable_variables():
        tf.summary.histogram(var.op.name, var)
    return train_op


def evaluation(logits, labels):
    """Evaluate the quality of the logits at predicting the label.

    Args:
        logits: Logits tensor, float - [batch_size, NUM_CLASSES].
        labels: Labels tensor, int32 - [batch_size], with values in the
        range [0, NUM_CLASSES).

    Returns:
        A scalar int32 tensor with the number of examples (out of batch_size)
        that were predicted correctly.
    """
    # For a classifier model, we can use the in_top_k Op.
    # It returns a bool tensor with shape [batch_size] that is true for
    # the examples where the label's is was in the top k (here k=1)
    # of all logits for that example.
    correct = tf.nn.in_top_k(logits, labels, 1)
    # Return the number of true entries.
    return tf.reduce_sum(tf.cast(correct, tf.int32))



"""
Observers to keep an eye on kids.

This model contains classes that are used to visualize network.
"""
from __future__ import absolute_import, division, print_function

import os
import sys
import inspect

import matplotlib as mpl
from six.moves import range
from six.moves import zip
mpl.use('Agg')
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d.art3d as art3d
import numpy as np
import math
import seaborn as sns
import tensorflow as tf
from pandas import Series, DataFrame

from .common import (
    ACTIVATION_COLLECTION,
    AUXILLIARY_STAT_COLLECTION,
    SPARSITY_SUMMARY_SUFFIX
)
from ..layers.synapse_layers import (
    ConvolutionLayer,
    InnerProductLayer,
    SynapseLayer,
)
from ..layers.activation_layers import ReLULayer
from .sensors import FeedSensor
from ..utils import glog as log, tf_to_caffe
from .. import backend as A

# Big figure size for visualization large amount of filters.
_FIG_SIZE = (100, 100)


class Observer(object):
    """
    An observer to visualize any trained brains.

    It takes the folder that saves the checkpoints(saved by tf.train.Saver) of
    a trained brain, as input to do visualization. It supports

    * activation visualization
    * filter visualization
    * 3D visualization of a single filter
    * statistical plotting on filters norm

    It also holds some general plotting routines which does not need a
    `Kid` to use. In that case, `kid` in the constructor could be None.

    Call APIs to draw any of those you want. Visualization will be saved at the
    model dir of the brain your has passed in.

    """
    def __init__(self, kid=None):
        """
        Args:
            kid: akid.core.kids.kid
                The kid whose brain is going to be observed.
        """
        self.kid = kid

    def visualize_classifying(self, name, idxs=None):
        """
        Visualize the prediction of a batch of samples from validation set.

        Args:
            name: str
                The name of the layer that whose `data` property holds the
                desired prediction, which could be logit or prediction
                probability.
            idxs: list
                Indices of images to be displayed. If not given, plot all
                images in the batch.

        """
        try:
            self._maybe_setup_kid()

            with self.kid.sess as sess:
                # Find the layer for logit or probability by name
                self.kid.restore_from_ckpt()
                feed_dict = self._feed_data(sess, get_val=True)
                data = self.kid.engine.get_layer_data(name, get_val=True)

                # Get the prediction for a batch.
                if type(self.kid.sensor) is FeedSensor:
                    _pred = data.eval(feed_dict=feed_dict)
                    data = feed_dict[self.kid.sensor.data(get_val=True)]
                    _ = feed_dict[self.kid.sensor.labels(get_val=True)]
                    if type(_) is list:
                        labels = _[0]
                    else:
                        labels = _
                else:
                    _pred, data, labels = sess.run([
                        data,
                        self.kid.sensor.data(get_val=True),
                        self.kid.sensor.labels(get_val=True)[0]])
                assert (len(_pred.shape) is 2,
                        """
                        Prediction result should be of shape (N, P), while
                        being {}, where N is the batch size, P the logit or
                        probability.
                        """.format(_pred.shape))
                pred = np.argmax(_pred, 1)

                # Plot images alongside labels.
                batch_size = pred.shape[0]
                is_color = data.shape[-1] == 3
                if is_color:
                    cmap = "cubehelix"
                else:
                    cmap = "gray"

                if idxs is None:
                    log.info("No indices of images are given. Will use the"
                             " all batch")
                    idxs = list(range(0, batch_size))

                # Initialize canvas.
                # Arrange image in a grid form as square as possible.
                img_num = len(idxs)
                cols = np.ceil(np.sqrt(img_num))
                rows = np.ceil(img_num / cols)
                log.info("Rows: {}; Cols: {};".format(rows, cols))
                fig = plt.figure()

                # Squeeze if grey image so it could be plotted.
                if data.shape[-1] is 1:
                    data = np.squeeze(data)
                # Plot images.
                log.info("Data will be display in {} row X {} col.".format(
                    rows, cols))
                for idx in idxs:
                    (img, label, pred_label) \
                        = (data[idx], labels[idx], pred[idx])
                    # Subplot No starts at 1. We add one to the idx.
                    axe = fig.add_subplot(rows, cols, idx+1)
                    axe.axis("off")
                    axe.set_title("Label: {}; PL: {}.".format(
                        label, pred_label))
                    axe.imshow(img, cmap=cmap)

                # Configure and show images.
                # Use plt.show() so that even in command line it could display
                # images.
                plt.show()

        except tf.OpError as e:
            log.info("Tensorflow error when running: {}".format(e.message))
            sys.exit(0)

    def visualize_activation(self):
        """
        Load trained net, do one forward propagation, and save activations.

        See more at doc string of `Observer` class.
        """
        log.info("Begin to draw activation of {}".format(self.kid.brain.name))
        try:
            self._maybe_setup_kid()

            with self.kid.sess.as_default():
                self.kid.restore_from_ckpt()
                feed_dict = self._feed_data(self.kid.sess)

                for block in self.kid.brain.blocks:
                    if block.data is None:
                        continue
                    log.info("Drawing activation of layer {}.".format(
                        block.name))
                    batch_activation = block.data.eval(feed_dict=feed_dict)
                    # For now we only visualize the first idx.
                    activation = batch_activation[0]
                    self.kid.sess.run(self.kid.brain.eval,
                             feed_dict=feed_dict)
                    activations_img = self._tile_to_one_square(activation)
                    title = "{} Layer".format(block.name)
                    # Visualization will be saved to an sub-folder of
                    # self.kid.model_dir
                    visualization_dir = self.kid.model_dir + "/visualization"
                    if not os.path.exists(visualization_dir):
                        os.mkdir(visualization_dir)
                    filename = visualization_dir + '/' + \
                        block.name + "_fmap.jpg"
                    self._heatmap_to_file(activations_img, title, filename)
        except tf.OpError as e:
            log.info("Tensorflow error when running: {}".format(e.message))
            sys.exit(0)

    def visualize_filters(self,
                          layers=None,
                          layout={"type": "normal", "num": 1},
                          big_resolution=True):
        """
        Load trained model and draw its filters to files.

        See more at doc string of `Observer` class.

        Args:
            layers: list
                If not None, only visualize filters of layers in the list.
            layout: dict
                How many columns of filters we should lay out.  It supports
                layouts: "square", "normal", "inverse", and "dynamic". "normal"
                and "inverse" support a parameter "num".

                "square" layout is for saving space, such as putting visualized
                filters on page, and is useful for plotting first layer
                filters, and inner product layers. If True, filters won't be
                organized by its 3D structure, but grouped into square to save
                space. Bias will not be drawn.

                The remaining layouts will layout filters preserving their 3D
                structure, which is to say to put 2D filters in a channel
                together, and also draw bias. Filters are organized by columns.

                "dynamic" will determined column number dynamically to make the
                final map as square as possible.

                "normal" layout will layout filters by specified column number
                with a further parameter `num`, which means the number of
                columns will the final filter map has. A larger column number
                corresponds to a fatter feature map. For example, if the value
                is 1, all filters will be in one column, which would result in
                a very thin image. If the column number cannot divide filter
                number, one more column will be used.

                "inverse" layout will layout filters by specified filters
                number in a column, with a further parameter, `num`, which
                means the number of filters in a column.
        """
        log.info("Begin to draw filters of {}".format(self.kid.brain.name))
        self._maybe_setup_kid()

        if A.backend() == A.TF:
            A.init(continue_from_chk_point=True, model_dir=self.kid.model_dir)

        for block in self.kid.brain.blocks:
            if layers:
                if block.name not in layers:
                    continue
            if not issubclass(type(block), SynapseLayer):
                continue
            log.info("Begin to tile the filter of layer {}".format(
                block.name))
            filters_img = self._tile_filters(block,
                                             layout,
                                             padding=2)
            title = "{} Layer".format(block.name)
            # Visualization will be saved to an sub-folder of
            # self.kid.model_dir
            visualization_dir = self.kid.model_dir + "/visualization"
            if not os.path.exists(visualization_dir):
                os.mkdir(visualization_dir)
            filename = visualization_dir + '/' \
                + block.name + "_para.png"
            self._heatmap_to_file(
                filters_img,
                title,
                filename,
                fig_size=_FIG_SIZE if big_resolution else None)

    def do_stat_on_filters(self):
        """
        Load trained model and do statistics on filters.

        See more at doc string of `Observer` class.
        """
        log.info("Begin to do stat on filters of {}".format(
            self.kid.brain.name))
        try:
            self._maybe_setup_kid()

            with self.kid.sess as sess:
                self.kid.restore_from_ckpt()

                stat_ops = tf.get_collection(AUXILLIARY_STAT_COLLECTION)
                stats = sess.run(stat_ops)

                # Draw histogram to file.
                for i, s in enumerate(stats):
                    # Plot.
                    fig = plt.figure()
                    axe = fig.add_subplot(111)
                    sns.set_context("poster")
                    sns.distplot(s, kde=False, rug=True)

                    # Save to file.
                    name = stat_ops[i].op.name.replace('/', '_')
                    axe.set_title(name)
                    # Visualization will be saved to an sub-folder of
                    # self.kid.model_dir
                    visualization_dir = self.kid.model_dir + "/visualization"
                    if not os.path.exists(visualization_dir):
                        os.mkdir(visualization_dir)
                    filename = visualization_dir + '/' + \
                        name + ".jpg"
                    fig.savefig(filename)
                    # We reuse title as a description of the file.
                    log.info("{} saved to {}.".format(name, filename))

        except tf.OpError as e:
            log.info("Tensorflow error when running: {}".format(e.message))
            sys.exit(0)

    def stem3_data(self):
        log.info("Stem3D output data of sensor `{}`".format(
            self.kid.sensor.name))

        try:
            if not self.kid.sensor.is_setup:
                self.kid.sensor.forward()

            with tf.Session(graph=self.kid.graph) as sess:

                if type(self.kid.sensor) is FeedSensor:
                    # Placeholder of `FeedSensor` should be filled.
                    feed_dict = self.kid.sensor.fill_feed_dict()
                    data_batch = feed_dict[self.kid.sensor.data()]
                else:
                    data_batch = self.kid.sensor.data().eval(session=sess)

                # For now we only use the first idx.
                data = data_batch[0]
                shape = data.shape
                # We only deal with gray level image now.
                data = data.reshape(shape[:-1])
                # TODO(Shuai): Deal with color image.
                self._stem3(data)
        except tf.OpError as e:
            log.info("Tensorflow error when running: {}".format(e.message))
            sys.exit(0)

    def _maybe_setup_kid(self):
        if not self.kid.brain.is_setup:
            # Do not do summary during visualization so we do not need to
            # create useless event files.
            self.kid.do_summary = False

            if A.backend() == A.TF:
                A.restore(self.kid.model_dir)

            self.kid.setup()

    def _feed_data(self, sess, get_val=False):
        if type(self.kid.sensor) is FeedSensor:
            # Placeholder of `FeedSensor` should be filled.
            feed_dict = self.kid.sensor.fill_feed_dict(get_val=get_val)
        else:
            feed_dict = None
            tf.train.start_queue_runners(sess=sess)

        return feed_dict

    def _heatmap_to_file(self, img,  title, filename, fig_size=None, **kwargs):
        """
        Given a image, draw it as heat map and save it to file.

        Parameters
        ----------
        img : numpy.array
            A heat map in form of 2D numpy array.
        title: str
            The title that will be appended at the top of the heat map.
        filename: str
            File name of the saved file.
        Remaining keywords will be passed to seaborn's `heatmap` function
        directly.
        """
        fig = plt.figure(figsize=fig_size)
        axe = fig.add_subplot(111)
        sns.set_context("poster")
        if img.min() >= 0:
            # If we know the date are non-negative, we tell seaborn so it could
            # use better visualization color map.
            sns.heatmap(data=img,
                        vmin=0, vmax=img.max(),
                        xticklabels=False, yticklabels=False,
                        square=True, robust=True,
                        **kwargs)
        else:
            sns.heatmap(data=img, center=0,
                        xticklabels=False, yticklabels=False,
                        square=True,
                        **kwargs)
        axe.set_title(title)
        fig.savefig(filename)
        # We reuse title as a description of the file.
        log.info("{} saved to {}.".format(title, filename))

    def _dataframe_to_file(self, df, title, filename):
        """
        Given an data frame whose columns are data that are supposed to be
        plotted as line graphs in one figure, save it to file.
        """
        fig = plt.figure()
        ax = fig.add_subplot(111)
        df.plot(title=title, ax=ax)
        sns.set_context("poster")
        ax.set_title(title)
        fig.savefig(filename)
        # We reuse title as a description of the file.
        log.info("{} saved to {}.".format(title, filename))

    def _get_a_canvas(self, height, width):
        """
        Get a background canvas for drawing images.

        We separate this out basically because we could centralize the thing
        related to background image, so change would be easier, such as
        changing background from white to black to make the boundary more
        obvious.

        Parameters
        ----------
        height : int
            Height of the canvas.
        width : int
            Width of the canvas

        Returns
        -------
        canvas : numpy.array
            A background image matrix.
        """
        return np.ones((height, width)) * 0

    def _get_shape_to_draw(self, filter_weights, name):
        """
        Get the actual shape we want to use to draw filter weights.

        It mainly deal with the case of inner product layer. When the previous
        inner product layer is a spatial feature map, it is better to keep the
        spatial information. In other cases, the returned shape is just the
        weights' original shape.

        WARNING: it is the legacy code to visualize Caffe models. It is not
        used anymore.

        Args:
            filter_weights: numpy.array
                Filter weights we want to visualize.
            name: str
                The name of the layer. It is used to locate the layer in the
                net's list, so previous layer could be fetched to determine the
                shape to draw for inner product layer.
        Returns:
            shape_to_draw: list
                shape used to draw the weights of this layer.
        """
        shape_to_draw = []
        if len(filter_weights.shape) is 2:
            # Means we are dealing with a inner product layer. If the previous
            # layer of this inner product layer is a convolution layer, we
            # display weights of inner product layer in the layout of last
            # layer's feature map so spatial information is kept.

            # Get the shape of last layer's feature map.
            layer_idx = self.kid.brain.block.index(name)
            previous_layer = self.kid.brain.blocks[layer_idx - 1]
            shape = previous_layer.data.get_shape().as_list()
            if len(shape) is 4:
                shape_to_draw.append(shape[1])
                shape_to_draw.append(shape[2])
                shape_to_draw.append(shape[3])
                shape_to_draw.append(filter_weights.shape[1])
                shape_to_draw = tuple(shape_to_draw)
            elif len(shape) is 2:
                shape_to_draw = filter_weights.shape
        else:
            # means we are dealing with a conv layer.
            shape_to_draw = filter_weights.shape

        return shape_to_draw

    def _draw_4D_filter(self,
                        filter,
                        layout={"type": "normal", "num": 2},
                        padding=1):
        """
        Draw 4D filters in multiple columns without biases.
        """
        w = filter[0]
        b = filter[1]
        w_height, w_width, channel_num, filter_num = w.shape

        pad_h = padding
        pad_w = padding
        w_width_pad = w_width + pad_w
        _w_height = w_height + pad_h
        _w_width = w_width_pad * channel_num
        _w_width += pad_w + 1  # For bias

        _col_num = None
        if layout["type"] == "dynamic":
            # First get the column number to handle most of the filters
            _col_num = int(math.floor(
                (filter_num * _w_height / _w_width) ** (0.5)))
            # Handle the special case we only need one column. It cannot be
            # absorbed in the special case because we need compute filters per
            # column using the un-incremented value first, in which case
            # division by 0 error would happen.
            if _col_num is 0:
                _col_num = 1
                filters_per_col = filter_num // _col_num
            else:
                filters_per_col = filter_num // _col_num
                # Deal with remaining filters if there is any
                if filter_num % _col_num is not 0:
                    _col_num += 1
            log.info("Column number is dynamically determined as {}".format(
                _col_num))
        elif layout["type"] == "inverse":
            filters_per_col = layout["num"]
            _col_num = int(np.ceil(filter_num / filters_per_col))
        elif layout["type"] == "normal":
            filters_per_col = filter_num // layout["num"]
            if filter_num % layout["num"] is not 0:
                # Append one more column to deal with the remaining filters.
                _col_num = layout["num"] + 1
            else:
                _col_num = layout["num"]
        else:
            log.info("Layout type `{}` is not supported. You perhaps have a"
                     " type".format(layout["type"]))
            sys.exit(1)

        canvas_h = _w_height * filters_per_col
        canvas_w = _w_width * _col_num
        out_img = self._get_a_canvas(canvas_h, canvas_w)
        log.info(
            "Final image's dimension is going to be {}.".format(out_img.shape))

        for f_idx in range(0, filter_num):
            # Flatten channels into one image
            filter_img = self._get_a_canvas(_w_height, _w_width)
            for c_idx in range(0, channel_num):
                col_idx = c_idx*w_width_pad
                filter_img[0:w_height, col_idx:col_idx+w_width] \
                    = w[..., c_idx, f_idx]
            # Add bias in.
            filter_img[0, -1] = b[f_idx]

            row_idx = (f_idx % filters_per_col) * _w_height
            col_idx = (f_idx // filters_per_col) * _w_width
            out_img[row_idx:row_idx+_w_height, col_idx:col_idx+_w_width] \
                = filter_img

        return out_img

    def _draw_4d_filter_one_col(self, w, b, padding=1):
        """
        Deprecated, and not used anymore. It is legacy code from days I still
        used Caffe.

        Draw 4D filters with bias on the right of a filter.
        """
        w_height, w_width, channel_num, filter_num \
            = w.shape
        # Transpose to Caffe style filter since the code below is original
        # written for Caffe.
        w = tf_to_caffe.tf_to_caffe_blob(w)

        # Traditional CNN only has one bias term per set of filters, we need to
        # how which case we are dealing with.
        if len(b.shape) == 1:
            log.info("Traditional CNN only has one bias term per filter set."
                     " Bias term will be drawn at the end of each row of"
                     " filter set")
            is_old_cnn = True
        else:
            log.info("Each filter of CNN has a bias term. The bias term will"
                     " be drawn side by side with the filter weight image.")
            is_old_cnn = False

        # Init output image size.
        # #################################################################
        # We do some pad between each image so we could distinguish them.
        pad_h = padding
        pad_w = padding
        # One filter consists of filter weight matrix and bias term.
        bias_w = 1  # We explicitly define bias width here for clarity.
        _w_height = w_height + pad_h
        if is_old_cnn:
            _w_width = w_width + pad_w
            out_img = self._get_a_canvas(_w_height * filter_num,
                                         # We add one for the bias term of one
                                         # set of filters
                                         _w_width * channel_num + 1)
        else:
            _w_width = w_width + bias_w + pad_w
            out_img = self._get_a_canvas(_w_height * filter_num,
                                         _w_width * channel_num)
        log.info(
            "Final image's dimension is going to be {}.".format(out_img.shape))

        # Tile the image
        # #################################################################
        for filter_idx in range(0, filter_num*channel_num):
            # Gather stats.
            filter_No = np.floor(filter_idx / channel_num)
            filter_channel_No = filter_idx % channel_num

            # Merge weight image and bias first.
            filter_img = self._get_a_canvas(_w_height, _w_width)
            filter_img[0:w_height, 0:w_width] \
                = w[filter_No, filter_channel_No, ...]
            # If it is not traditional CNN, we append bias term after each
            # filter.
            if not is_old_cnn:
                filter_img[_w_height, _w_width] \
                    = b[filter_No, filter_channel_No]

            # Tile weight image and bias in.
            row_idx = filter_No * _w_height
            col_idx = filter_channel_No * _w_width
            out_img[row_idx:(row_idx+_w_height), col_idx:(col_idx+_w_width)] \
                = filter_img

        # If it is traditional CNN, we append the bias term at the end of each
        # row of output image.,
        for filter_No in range(0, filter_num):
            out_img[filter_No*_w_height, channel_num*_w_width] \
                = b[filter_No]

        return out_img

    def _draw_2D_filter(self, filter):
        raise NotImplementedError("2D filter visualization is not implemented"
                                  " yet!")
        sys.exit()

    def _stem3(self, X, ax=None):
        """
        Draw an image X using 3D stem plot on the given axe object.

        Note that this routine will automatically do a horizontal mirror
        operation given the difference between the image coordinate system and
        the normal Euclidean system. So it does not work for normal stem3
        plotting.

        Parameters
        ----------
        X : numpy.array
            2D matrix to draw.
        ax : matplotlib.axes._subplots.Axes3DSubplot
            The axe object.
        """
        assert len(X.shape) is 2, \
            "Only gray level images are supported. The input's shape is"\
            " {}".format(X.shape)
        # Setup coordinates.
        h, w = X.shape
        x = np.linspace(0, w-1, w)
        y = np.linspace(0, h-1, h)
        xv, yv = np.meshgrid(x, y)

        # If no axis object is given, we need to create it here.
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1, projection='3d')
        # Draw.
        for row in range(h):
            for x_, y_, z_ in zip(xv[row, :], yv[row, :], X[h-1-row, :]):
                line = art3d.Line3D(*list(zip((x_, y_, 0), (x_, y_, z_))),
                                    marker='o')
                ax.add_line(line)
        ax.set_xlim3d(0, w)
        ax.set_ylim3d(0, h)
        ax.set_zlim3d(min(0, X.min()), X.max() * 1.3)

        plt.show()

    def _tile_to_one_square(self, images):
        """
        Tile a set of images into one image.

        Given that there is no sub-level group in images, we tile it to an
        image as square as possible to save space.

        To be able to zoom in and move around to look into the filtered
        results, in the case that we do not need the label information as
        inspecting input data, all filtered results are better to be displayed
        in one image.

        Parameters
        ----------
        images : numpy.array
            The set of images with dimension (H, W, C), where the C is the
            channel number the set of images.

        Returns
        -------
        out_img : numpy.array
            The 2D image tiled with input images.
        """
        log.info("Tiling 3D images to 2D with square layout.")
        if len(images.shape) is 1:
            # means we are dealing with inner product layer.
            channel_num = images.shape[0]
            height = width = 1
            # No padding for ip layer, since each channel only has one pixel.
            pad_h = 0
            pad_w = 0
        else:
            # means we are dealing with conv layer.
            # Convert from tf style to caffe style. (We do this conversion
            # since below is the code that works with Caffe, which I wrote in
            # the days of using Caffe.)
            images = tf_to_caffe.tf_to_caffe_blob(images)
            (channel_num, height, width) = images.shape
            # We do some pad between each image so we could distinguish them.
            pad_h = 1
            pad_w = 1
        # Calculate the size of the final image.
        # Arrange image in a grid form as square as possible.
        cols = int(np.ceil(np.sqrt(channel_num)))
        rows = int(np.ceil(channel_num / cols))
        pad_h = height + pad_h
        pad_w = width + pad_w
        out_img_h = pad_h * rows
        out_img_w = pad_w * cols
        out_img = self._get_a_canvas(out_img_h, out_img_w)
        # Fill the feature map in the output image.
        for map_idx in range(0, channel_num):
            row_idx = int(np.floor(map_idx / cols))
            col_idx = map_idx % cols
            x = row_idx * pad_h
            y = col_idx * pad_w
            out_img[x:(x+height), y:(y+width)] \
                = images[map_idx, ...]
        log.info("Tiling ends.")
        return out_img

    def _tile_filters(self,
                      block,
                      layout,
                      padding=1):
        """
        Tile filters of one layer into one image.

        To be able to compare filters applied on the same feature map, One set
        of filters is going to be grouped in one line. Bias will be side by
        side with the filter weight image.

        Parameters
        ----------
        block: akid.layers.ConnectionLayer
            Which block's filters to visualize.
        layout: an dict
            See method `visualize_filters`.

        Returns
        -------
        out_img : numpy.array
            The tiled image.
        """
        block_type = type(block)

        assert issubclass(block_type, ConvolutionLayer) or\
            issubclass(block_type, InnerProductLayer), \
            "Block type {} is not supported!".format(block_type)

        ret = A.eval(block.var_list)
        w, b = ret[0], ret[1]
        if A.backend() == A.TORCH:
            if len(w.shape) == 4:
                w = np.einsum('{}->{}'.format('oihw', 'hwio'), w)
        if block.bag:
            if "filters_to_visual" in block.bag:
                filters_to_visual = block.bag["filters_to_visual"]
                w = w[..., filters_to_visual]
                b = b[filters_to_visual]

        try:
            layout_type = layout["type"]
        except KeyError as e:
            raise KeyError("Error: {}. Layout `type` is not found in layout: {}. A"
                           " layout type is needed to determine how to draw"
                           " filters. You perhaps have a typo".format(e.message,
                                                                      layout))

        if layout_type is "square":
            shape = w.shape
            if len(shape) is 2:
                # it's a inner product layer.
                w = w.reshape([-1])
                filter_img = self._tile_to_one_square(w)
            else:
                w = w.reshape([shape[0], shape[1], -1])
                filter_img = self._tile_to_one_square(w)
        else:
            if issubclass(block_type, InnerProductLayer):

                # We do not want to modify block.in_shape's value. So make a
                # copy of it.
                if len(block.in_shape) == 3:
                    if A.backend() == A.TORCH:
                        c_in, h_in, w_in = block.in_shape[0], block.in_shape[1], block.in_shape[2]
                        shape_to_draw = [h_in, w_in, c_in]
                else:
                    shape_to_draw = [i for i in block.in_shape]
                shape_to_draw.append(w.shape[-1])

                # If previous layer is IP layer, we just square tile it.
                if len(shape_to_draw) is 2:
                    w = w.reshape([-1])
                    filter_img = self._tile_to_one_square(w)
                else:
                    # else we reshape it to the shape of last feature map for
                    # better visualization.
                    log.info("Inner product layer {} is reshaped to the shape"
                             " of its previous layer to visualize: {} ->"
                             " {}.".format(
                                 block.name, block.shape, shape_to_draw))
                    w = w.reshape(shape_to_draw)

                    filter_img = self._draw_4D_filter([w, b],
                                                      layout=layout,
                                                      padding=1)
            else:
                filter_img = self._draw_4D_filter([w, b],
                                                  layout=layout,
                                                  padding=1)

        return filter_img

    def get_filters(self, model_dir):
        """
        Load trained model from `model_dir`.

        This is a legacy code from the days when I was visualizing Caffe
        trained model. In that case, I do not have easy access to network
        topology, so have to hack to get all filters. `self.visualize_filters`
        now does not use this method anymore.

        Returns:
            filter_dict: dict
                Filters are returned in form of dictionary:
                {
                    name: value
                    ...
                }
                name is layer name with type str while value is a numpy array.
        """
        filter_dict = {}
        with tf.Session(graph=self.kid.graph) as sess:
            if self.kid.brain.infer_graph is None:
                self.kid.setup(sess, continue_from_chk_point=True)
            filters = self.kid.val_brain.get_filters()
            # Eval all parameters.
            # We use an explicit loop since we need both weights and biases
            # at the same time.
            for i in range(0, len(filters), 2):
                weights_tensor = filters[i]
                biases_tensor = filters[i+1]
                name = weights_tensor.name.split('/')[0]
                weight_array = weights_tensor.eval()
                biases_array = biases_tensor.eval()
                filter_dict[name] = [weight_array, biases_array]
        return filter_dict

    def get_activation(self, model_dir, idxs=[0]):
        """
        Notice:

        This is a obsolete method that is used at days when I just ported the
        code from visualization code for Caffe. It is not used now.

        Load trained model from `model_dir`.
        Activation are returned in form of dictionary:
        {
            name: value
            ...
        }
        name is a str while value is a numpy array.

        Args:
            model_dir: str
                where the trained net stored.
            idxs: list of integer
                Normally, the net returns activations in batch. idxs is the
                indices of activations we want. By default, only the first
                activation is returned.
        """
        activation_dict = {}
        with tf.Session(graph=self.kid.graph) as sess:
            if self.kid.brain.infer_graph is None:
                self.kid.brain._setup_infer_graph()
            # Recover inference part of the net.
            activations = tf.get_collection(ACTIVATION_COLLECTION)
            # Remove data activation from the list, given we have to feed the
            # value to it.
            for a in activations:
                if a.name.find("data") > -1:
                    # Save data tensor so it could be used to get data values
                    # later.
                    data = a
                    activations.remove(a)
            # Init.
            saver = tf.train.Saver(tf.global_variables())
            self.restore_from_ckpt(sess, saver, model_dir)
            # Run the graph once to get all activation.
            feed_dict = self.fill_feed_dict(self.data_sets.train)
            activation_values = sess.run(activations, feed_dict=feed_dict)
            # Add data to activation_dict so it could also be visualized.
            if len(idxs) is 1:
                activation_dict["data"] = feed_dict[data][idxs[0], ...]
            else:
                activation_dict["data"] = feed_dict[data][idxs, ...]
            # Add actual activation.
            for i in range(0, len(activations)):
                # Shorten the name.
                name = activations[i].name.split('/')[0]
                if len(idxs) is 1:
                    activation_dict[name] = activation_values[i][idxs[0], ...]
                else:
                    activation_dict[name] = activation_values[i][idxs, ...]
        return activation_dict

    def plot_relu_sparsity(self, EVENT_FILE_PATH):
        log.info("Begin to plot relu sparsity curve of {}".format(
            self.kid.brain.name))
        try:
            if not self.kid.brain.is_setup:
                # Do not do summary during visualization so we do not need to
                # create useless event files.
                self.kid.do_summary = False
                self.kid.setup()

            # Gather tag names for sparsity of relu layers.
            relu_sparsity_tag_list = []
            # The list holds name for columns of final data frame, which would
            # be used as legend name is the produced plot.
            dt_col_name_list = []
            for block in self.kid.brain.blocks:
                if type(block) is ReLULayer:
                    relu_sparsity_tag_list.append(
                        block.data.op.name + '/' + SPARSITY_SUMMARY_SUFFIX)
                    dt_col_name_list.append(block.data.op.name.split('/')[-2])

            # Set up a dictionary to hold sparsity data
            relu_sparsity_dict = {}
            for name in dt_col_name_list:
                relu_sparsity_dict[name] = Series()

            log.info("Gathering sparsity info of {}.".format(
                self.kid.brain.name))
            for e in tf.train.summary_iterator(EVENT_FILE_PATH):
                for v in e.summary.value:
                    for i, tag in enumerate(relu_sparsity_tag_list):
                        if v.tag == tag:
                            relu_sparsity_dict[dt_col_name_list[i]].set_value(
                                e.step, v.simple_value)

            sparsity_df = DataFrame(relu_sparsity_dict)
            title = "Activation sparsity of different layers after ReLU."
            # Visualization will be saved to an sub-folder of
            # self.kid.model_dir
            visualization_dir = self.kid.model_dir + "/visualization"
            if not os.path.exists(visualization_dir):
                os.mkdir(visualization_dir)
            filename = visualization_dir + "/relu_sparsity.jpg"
            self._dataframe_to_file(sparsity_df, title, filename)
        except tf.OpError as e:
            log.info("Error when reading event file: {}".format(e.message))
            sys.exit(0)

    def line_plot(self,
                  x_axis,
                  y_axis,
                  x_label,
                  y_label,
                  title,
                  log_x=False):
        """
        Given a list of data, and corresponding meta info, plot a line graph
        using it.

        Args:
            x_axis: list
                The list contains the values of y axis.
            y_axis: list
                The list contains the values of x axis.
            log_x: Boolean
                If True, x_axis will be taken log.
        """
        series = Series(data=y_axis, index=x_axis)

        # Plot
        sns.set_context("talk")
        fig = plt.figure()
        axe = fig.add_subplot(111)
        axe.set_xlabel(x_label)
        axe.set_ylabel(y_label)
        axe.set_title(title)
        series.plot(ax=axe, kind="line", logx=log_x)
        filename = "{}.jpg".format(title.replace(' ', '_'))
        fig.savefig(filename)

__all__ = [name for name, x in locals().items() if
           not inspect.ismodule(x) and not inspect.isabstract(x)]



from __future__ import division

from __future__ import absolute_import
import abc
import inspect

import tensorflow as tf
from deprecated import deprecated

from ..core.blocks import ProcessingLayer
from ..core.common import (
    AUXILLIARY_SUMMARY_COLLECTION,
    AUXILLIARY_STAT_COLLECTION
)
from .. import backend as A
from .activation_layers import BatchNormalizationLayer
from .. import nn

from torch.nn import BatchNorm2d
from six.moves import range


class SynapseLayer(ProcessingLayer):
    """
    An abstract class to model connections between neurons. It takes massive
    inputs, which are dendrites in neuroscience, and produces massive raw
    outputs, which are going to be further processed by consequent layers, such
    as pooling layers and non-linear activation layers. The outputs ultimately
    connect to next layer, which is axons in neuroscience. Convolution and
    Inner Product layers are two examples.

    A large number of connections needs to be initialized before training
    starts. Constructor parameter `initializer` lets user pass in the type of
    initializer one wants. Current support initializers are:

        * truncated_normal.
              Customizable parameters: stddev, standard deviation of the normal
              distribution; means are zero by default.
        * uniform
              Customizable parameters: range; A uniform initializer with range
              1 will have uniform distribution U(-1, 1).
    """
    def __init__(self,
                 in_channel_num,
                 out_channel_num,
                 initial_bias_value=0,
                 init_para={"name": "default"},
                 max_norm=None,
                 do_stat_on_norm=False,
                 # wd={"type": "l2", "scale": 5e-4},  # Save for future reference.
                 wd=None,
                 wd_on_bias=False,
                 **kwargs):
        """
        Args:
            in_channel_num: int
                How many in channels this synapse layer has.
            out_channel_num: int
                How many out channels this synapse layer has.
            init_bias_value: a real number
                If bias is unwanted, set it to None explicitly.
            init_para: dict
                An dictionary that contains the name of the initializer to use,
                and its parameters. The name, and parameters have to exactly
                match with perdefined strings.
            max_norm: a real number
                A real number that constrains the maximum norm of the filter of
                a channel could be at largest. If exceeding that value, it
                would be projected back.
            do_stat_on_norm: boolean
                Whether to collect statistics on norms of filters. If true,
                summary ops that calculate norms will be added to
                `AUXILIARY_SUMMARY_COLLECTION`, which could be used further. It
                is mainly used for quantitatively evaluate how many filters are
                dead during training.
            wd: dict
                An dictionary that contains the `type` of regularization to use
                and its scale, which is to say the multiplier of the
                regularization loss term. If None, weight decay is not added.
            wd_on_bias: boolean
                Whether to apply weight decay on biases.
        """
        super(SynapseLayer, self).__init__(**kwargs)
        self.in_channel_num = in_channel_num
        self.out_channel_num = out_channel_num
        self.init_para = init_para
        self.max_norm = max_norm
        self.do_stat_on_norm = do_stat_on_norm
        self.wd = wd
        self.wd_on_bias = wd_on_bias

        # Only do float conversion if not None.
        self.initial_bias_value = float(initial_bias_value) \
            if initial_bias_value is not None else None

    def _post_setup(self):
        super(SynapseLayer, self)._post_setup()
        if self.weights is not None:
            self.log("Weight shape {}".format(A.get_shape(self.weights)))

        if self.initial_bias_value is None:
            self.log("Bias is disabled.")
        else:
            # It is possible some layers does not use bias.
            if self.biases is not None:
                self.log("Bias shape {}".format(A.get_shape(self.biases)))

        if self.wd is not None:
            self.log("Using {} regularization with scale {}".format(
                self.wd["type"], self.wd["scale"]))

            if self.wd_on_bias:
                self.log("Apply regularizer {} with scale {} on bias.".format(
                    self.wd["type"], self.wd["scale"]))


        if self.do_stat_on_norm:
            for v in self.var_list:
                # Do not apply on biases.
                shape = v.get_shape().as_list()
                if len(shape) > 1:
                    _ = tf.reduce_sum(tf.square(v), list(range(0, len(shape)-1)))
                    v_norms = tf.sqrt(_, name="l2norm")
                    tf.add_to_collection(AUXILLIARY_STAT_COLLECTION,
                                         v_norms)
                    tf.summary.histogram(
                        v.op.name + "l2norm",
                        v_norms,
                        collections=[AUXILLIARY_SUMMARY_COLLECTION])

    def _pre_forward(self, *args, **kwargs):
        super(SynapseLayer, self)._pre_forward(*args, **kwargs)
        if self.wd and self.wd["scale"] is not 0:
            self._loss = nn.regularizers.compute(self.wd["type"],
                                                 var=self.weights,
                                                 scale=self.wd["scale"],
                                                 name='weight_loss' if self.summarize_output else None)

        if self.initial_bias_value is not None:
            if self.wd_on_bias and self.wd["scale"] is not 0:
                loss = nn.regularizers.compute(self.wd["type"],
                                               var=self.biases,
                                               scale=self.wd["scale"],
                                               name='bias_loss' if self.summarize_output else None)

                if self._loss is not None:
                    self._loss += loss
                else:
                    self._loss = loss


    def on_para_update(self):
        if self.max_norm:
            self.log("Using max norm constrain of {}.".format(self.max_norm))
            # Create the op to apply max norm clip on weights.
            self.clipped_filters = []
            for v in self.var_list:
                # Do not apply on biases.
                shape = v.get_shape().as_list()
                if len(shape) > 1:
                    out_channel_dim = len(shape) - 1
                    filter_tuple = tf.split(axis=out_channel_dim, num_or_size_splits=shape[-1], value=v)
                    clipped_filter_list = []
                    for f in filter_tuple:
                        clipped_filter_list.append(
                            tf.clip_by_norm(f, self.max_norm))
                    clipped_v = tf.concat(axis=out_channel_dim,
                                          values=clipped_filter_list)
                    self.clipped_filters.append(tf.assign(v, clipped_v))
        else:
            self.clipped_filters = []

        return self.clipped_filters

    def get_clone(self):
        c = super(SynapseLayer, self).get_copy()

        # TODO: current clone after setup does not have names.
        if self.is_setup:
            c.weights = A.Tensor(A.value(c.weights), requires_grad=True)
            c.var_list = [c.weights]
            if c.biases is not None:
                c.biases = A.Tensor(A.value(c.biases), requires_grad=True)
                c.var_list.append(c.biases)

        return c

    @property
    def weights(self):
        return self._weights

    @weights.setter
    def weights(self, v):
        self._weights = v

    @property
    def biases(self):
        return self._biases

    @biases.setter
    def biases(self, v):
        self._biases = v


class ConvolutionLayer(SynapseLayer):
    NAME = "Conv2D"
    def __init__(self,
                 ksize,
                 strides=1,
                 padding="SAME",
                 depthwise=False,
                 bn=False,
                 **kwargs):
        super(ConvolutionLayer, self).__init__(**kwargs)
        self.ksize = [ksize, ksize] if type(ksize) is int else ksize

        self.strides = strides
        if A.backend() == A.TORCH:
            if type(strides) is int:
                self.strides = [strides, strides]
            elif len(strides) == 4:
                # The stride format is of that of tensorflow format. Extract the
                # stride to torch format.
                self.strides = [strides[1], strides[2]]

        self.padding = padding
        self.depthwise = depthwise
        self.bn = bn

    def get_weigth_shape(self):
        if A.backend() == A.TORCH:
            if len(self.ksize) == 4:
                # The ksize format is of that of tensorflow format. Extract the
                # ksize to torch format.
                self.ksize = [self.ksize[1], self.ksize[2]]
            shape = [self.out_channel_num, self.in_channel_num,
                          self.ksize[0], self.ksize[1]]
        else:
            shape = [self.ksize[0], self.ksize[1],
                          self.in_channel_num, self.out_channel_num]

        return shape

    def _setup(self):
        self.shape = self.get_weigth_shape()
        self._weights = self._get_variable("weights",
                                          self.shape,
                                          self._get_initializer(self.init_para))

        if self.initial_bias_value is not None:
            self._biases = self._get_variable(
                'biases',
                [self.out_channel_num if not self.depthwise else self.out_channel_num * self.in_channel_num],
                initializer=self._get_initializer(init_para={"name": "constant",
                                                             "value": self.initial_bias_value}))
        else:
            self._biases = None

        self.log("Padding method {}.".format(self.padding), debug=True)

        # Emergency code. Should be cleansed in the future.
        if self.bn:
            self.bn = BatchNorm2d(self.out_channel_num, affine=False)
            self.bn = self.bn.cuda() if A.use_cuda() else self.bn

    def _pre_forward(self, input, *args, **kwargs):
        super(ConvolutionLayer, self)._pre_forward(*args, **kwargs)
        self.input_shape = A.get_shape(input)

    def _forward(self, input):
        name = 'fmap' if self.summarize_output else None

        if self.depthwise:
            conv = A.nn.depthwise_conv2d(input,
                                         self.weights,
                                         self.biases,
                                         self.strides,
                                         self.padding,
                                         name=name)
        else:
            conv = A.nn.conv2d(input, self.weights,
                               self.biases, self.strides,
                               self.padding, name=name)

        self._data = conv

        # Emergency code. Should be cleansed in the future.
        if self.bn:
            self._data = self.bn(self._data)

        return self._data

    @deprecated(reason="Legacy code.")
    def _on_update_stub(self, K_prev):
        # Save the metric for this RKHS
        self.metric = K_prev
        # Compute the metric for the next layer

        if self.bn:
            W = A.nn.normalize_weight(self.weights)
        else:
            W = self.weights
        self.K = A.nn.nn_riemannic_metric(K_prev, W, self.biases if not self.bn else None)

    def backward(self, X_in):
        """
        According to the top-down inference results, reconstruct the input.
        """
        # Deconvolve.
        if self.initial_bias_value is not None:
            X_in = tf.nn.bias_add(X_in, -self.biases)

        self._data_g = tf.nn.conv2d_transpose(X_in,
                                              self.weights,
                                              self.input_shape,
                                              self.strides,
                                              self.padding,
                                              name="deconv")

        # TODO: Create reconstruction loss.

        return self.data_g


class SelfNormalizedConvolutionLayer(SynapseLayer):
    NAME = "SNConv2D"
    def __init__(self,
                 ksize,
                 strides=1,
                 padding="SAME",
                 depthwise=False,
                 **kwargs):
        super(SelfNormalizedConvolutionLayer, self).__init__(**kwargs)
        self.ksize = [ksize, ksize] if type(ksize) is int else ksize

        self.strides = strides
        if A.backend() == A.TORCH:
            if len(strides) == 4:
                # The stride format is of that of tensorflow format. Extract the
                # stride to torch format.
                self.strides = [strides[1], strides[2]]

        self.padding = padding
        self.depthwise = depthwise

    def get_weigth_shape(self):
        if A.backend() == A.TORCH:
            if len(self.ksize) == 4:
                # The ksize format is of that of tensorflow format. Extract the
                # ksize to torch format.
                self.ksize = [self.ksize[1], self.ksize[2]]
            shape = [self.out_channel_num, self.in_channel_num,
                          self.ksize[0], self.ksize[1]]
        else:
            shape = [self.ksize[0], self.ksize[1],
                          self.in_channel_num, self.out_channel_num]

        return shape

    def _setup(self):
        self.shape = self.get_weigth_shape()
        self._weights = self._get_variable("weights",
                                          self.shape,
                                          self._get_initializer(self.init_para))

        if self.initial_bias_value is not None:
            self._biases = self._get_variable(
                'biases',
                [self.out_channel_num if not self.depthwise else self.out_channel_num * self.in_channel_num],
                initializer=self._get_initializer(init_para={"name": "constant",
                                                             "value": self.initial_bias_value}))
            if A.DATA_FORMAT == "CHW":
                self.biases = A.reshape(self.biases, [1, -1, 1, 1])
            elif A.DATA_FORMAT == "HWC":
                self.biases = A.reshape(self.biases, [1, 1, 1, -1])
            else:
                assert False, "Should not reach here. Data format is wrong."
        else:
            self._biases = None

        self.log("Padding method {}.".format(self.padding), debug=True)

    def _pre_forward(self, input, *args, **kwargs):
        super(SelfNormalizedConvolutionLayer, self)._pre_forward(*args, **kwargs)
        self.input_shape = A.get_shape(input)

    def _forward(self, input):
        name = 'fmap' if self.summarize_output else None

        conv = A.nn.conv2d(input, self.weights,
                           None, self.strides,
                           self.padding, name=name)

        self.mean = A.mean(conv, dim=[0, 2, 3], keep_dim=True)
        self.std = A.std(conv, dim=[0, 2, 3], keep_dim=True)
        out = (conv - self.mean) / self.std

        out = out + self.biases

        self._data = out

        return self._data


SNConv2D = SelfNormalizedConvolutionLayer


class Convolution1DLayer(SynapseLayer):
    NAME = "Conv1D"

    def __init__(self, ksize, stride, padding, **kwargs):
        super(Convolution1DLayer, self).__init__(**kwargs)
        self.ksize = ksize
        self.stride = stride
        self.padding = padding

    def _setup(self):
        if A.DATA_FORMAT == "CHW":
            self.shape = [self.out_channel_num, self.in_channel_num, self.ksize]
        else:
            raise ValueError("Data format {} is not supported yet.".format(A.DATA_FORMAT))
        self._weights = self._get_variable("weights",
                                          self.shape,
                                          self._get_initializer(self.init_para))
        if self.initial_bias_value is not None:
            self._biases = self._get_variable(
                'biases',
                [self.out_channel_num],
                 initializer=self._get_initializer(init_para={"name": "constant",
                                                              "value": self.initial_bias_value}))
        else:
            self._biases = None

    def _forward(self, x):
        self._data = A.nn.conv1d(x, self.weights, self.biases, self.stride, self.padding,
                                 name="fmap" if self.summarize_output else None)
        return self._data

Conv1D = Convolution1DLayer


class ColorfulConvLayer(ConvolutionLayer):
    def __init__(self,
                 c_W_initializer={"name": "uniform_unit_scaling", "factor": 1},
                 equivariant=False,
                 same_ksize=False,
                 verbose=False,
                 use_bn=True,
                 **kwargs):
        """
        Args:
            c_W_initializer: dict
                By default, uniform_unit_scaling_initializer is used. The rationale is
                to make the output variance (ideally should be norm) of color map and
                the feature map the same: the factor is set to 1 = 1/sqrt(3) *
                sqrt(COLOR_CHANNEL_NUM) --- COLOR_CHANNEL_NUM is 3. It is mostly for ad
                hoc experimental, which means this convolution layer should use default
                initializers.
            equivariant: bool
                Generate equivariant feature map if True, otherwise, use max inference.
            same_ksize: bool
                Use the same ksize with the filter for color filtering. If
                None, use 1 X 1 filter.
            use_bn: bool
                Use BN before addition. It is supposed to be used for testing,
                given it is harder to get the test case when BN is involved.
            verbose: bool
                If True, do histogram summary on both convolution map and color
                map.
        """
        super(ColorfulConvLayer, self).__init__(**kwargs)
        self.c_W_initializer = c_W_initializer
        self.verbose = verbose
        self.equivariant = equivariant
        self.same_ksize = same_ksize
        self.use_bn = use_bn

    def _pre_forward(self, input, *args, **kwargs):
        super(ConvolutionLayer, self)._pre_forward(*args, **kwargs)
        self.input_shape = input[0].get_shape().as_list()

    def _setup(self):
        super(ColorfulConvLayer, self)._setup()
        if self.same_ksize:
            shape = [self.ksize[0], self.ksize[1], 3, self.out_channel_num]
        else:
            shape = [1, 1, 3, self.out_channel_num]
        self.color_W, c_W_loss = self._variable_with_weight_decay(
            "color_weights", shape, self.c_W_initializer)
        if c_W_loss is not None:
            self._loss += c_W_loss

        if self.use_bn:
            self.color_bn = BatchNormalizationLayer(channel_num=self.out_channel_num, name="color_bn")
            self.shape_bn = BatchNormalizationLayer(channel_num=self.out_channel_num, name="shape_bn")
        else:
            self.color_bn = lambda x: x
            self.shape_bn = lambda x: x

    def _forward(self, X_in):
        F = X_in[0]
        C = X_in[1]
        F_out = super(ColorfulConvLayer, self)._forward(F)
        F_out = self.shape_bn(F_out)
        if self.verbose and not self.is_val:
            self._data_summary(F_out, sparsity_summary=False)

        C_out = A.nn.depthwise_conv2d(C, self.color_W, 1, self.padding)
        shape = C_out.get_shape().as_list()
        shape = [shape[0], shape[1], shape[2], 3, self.out_channel_num]
        C_out = A.reshape(C_out, shape)
        C_out = self.color_bn(C_out)
        if self.equivariant:
            F_out = A.expand_dims(F_out, axis=-2)
            out = F_out + C_out
            out = A.reshape(out, [shape[0], shape[1], shape[2], -1])

            if self.verbose and not self.is_val:
                self._data_summary(C_out, sparsity_summary=False)
                self._data_summary(A.abs(C_out)/(A.abs(F_out) + A.abs(C_out)), sparsity_summary=False)
        else:
            C_max = A.reduce_max(C_out, axis=3)

            if self.verbose and not self.is_val:
                self._data_summary(C_max)

            out = F_out + C_max

        self._data = out

        return self._data


class EquivariantProjectionLayer(SynapseLayer):
    def __init__(self, g_size, strides=1,  **kwargs):
        """
        NOTE: for this layer, the input should be g_size * in_channel_num, so
        is output.
        """
        super(EquivariantProjectionLayer, self).__init__(**kwargs)
        self.g_size = g_size
        self.strides = strides

    def _setup(self):
        self.weights, self._loss = self._variable_with_weight_decay(
            "projection_weights", [self.g_size, 1, 1, self.in_channel_num, self.out_channel_num], self.init_para)
        self.biases = None

    def _forward(self, X_in):
        shape = X_in.get_shape().as_list()
        new_shape = [shape[0], shape[1], shape[2], self.g_size, -1]
        X_in = A.reshape(X_in, new_shape)
        X_g = A.unstack(X_in, 3)
        X_out_g = []
        for i in range(self.g_size):
            X_out_g.append(
                A.nn.conv2d(X_g[i],
                             self.weights[i, ...],
                             strides=self.strides,
                             padding="SAME"))

        X_out = A.pack(X_out_g, axis=3)
        shape = X_out.get_shape().as_list()
        X_out = A.reshape(X_out, shape=[shape[0], shape[1], shape[2], -1])
        self._data = X_out

        return X_out


class SLUConvLayer(ConvolutionLayer):
    """
    Deprecated.

    This is a wrong idea. It tests sign after convolution, but it should be
    done before. Also, the sign test depends on the top down reconstruction
    value, which is a latent variable.

    Convolution layer preceded by Switchable Linear Unit, activation function
    developed by properly doing posterior inference in DRMM.
    """
    def _forward(self, X_in):
        depthwise = A.nn.depthwise_conv2d(X_in, self.weights, self.strides, self.padding)
        # Compute the sign of each pixel.
        sign = A.cast(depthwise > 0, A.float32)
        # Drop negative value.
        conv = depthwise * sign
        # Collapse dimensions.
        shape = conv.get_shape().as_list()
        shape.pop()
        shape.extend([self.out_channel_num, self.input_shape[-1]])
        conv = A.reshape(conv, shape=shape)
        conv = tf.reduce_sum(conv, axis=-1, name="slu_conv")

        if self.initial_bias_value is not None:
            output = tf.nn.bias_add(conv, self.biases, name="slu_conv_bias")
        else:
            output = conv

        self._data = output


class InnerProductLayer(SynapseLayer):
    NAME = "IP"

    def _forward(self, input):
        input = self._reshape(input)
        ip = A.nn.inner_product(input,
                                self.weights,
                                bias=self.biases,
                                name='fmap' if self.summarize_output else None)
        self._data = ip

        return self._data

    def _reshape(self, input):
        """
        Flatten all input feature maps.

        Args:
            input: tensor
                The processed input from previous layer, or just input.
        Return:
            input: tensor
                The reshaped input that could be processed by this layer.
        """
        input_shape = A.get_shape(input)
        in_channel_num = input_shape[1]
        # Check the input shape, if it is not 2D tensor, reshape all remaining
        # dimensions into one.
        reshaped_input = input
        if len(input_shape) != 2:
            in_channel_num = 1
            for i in input_shape[1:]:
                in_channel_num *= i
            flattened = A.reshape(input, [input_shape[0], in_channel_num])
            reshaped_input = flattened

        # Hold another addition info about the shape of input feature maps.
        self.in_shape = input_shape[1:]

        return reshaped_input

    def _setup(self):
        self.shape = [self.in_channel_num, self.out_channel_num]
        self._weights = self._get_variable(
            'weights', self.shape, self._get_initializer(self.init_para))

        if self.initial_bias_value is not None:
            self._biases = self._get_variable(
                'biases',
                shape=[self.out_channel_num],
                initializer=self._get_initializer({"name": "constant", "value": self.initial_bias_value}))
        else:
            self._biases = None

    @deprecated(reason="Legacy code.")
    def _on_update_stub(self, K_prev):
        # Save the metric for this RKHS
        self.metric = K_prev
        # Compute the metric for the next layer
        if len(self.in_shape) == 3:
            shape = self.in_shape[:]
            shape.insert(0, self.out_channel_num)
            W = A.reshape(self.weights, shape)
        else:
            W = self.weights

        self.K = A.nn.nn_riemannic_metric(K_prev, W, self.biases)


class InvariantInnerProductLayer(SynapseLayer):
    """
    This is a half-baked idea. See blog note for details.
    """
    NAME = "Invariant_IP"

    def _forward(self, input):
        object_vector = self._preprocess(input)
        ip = tf.matmul(object_vector, self.weights)
        ip_plus_bias = tf.nn.bias_add(ip,
                                      self.biases,
                                      name=InvariantInnerProductLayer.NAME)

        self._data = ip_plus_bias

    def _preprocess(self, input):
        # Gather some info.
        input_shape = input.get_shape().as_list()
        batch_size = input_shape[0]
        fmap_h = input_shape[1]
        fmap_w = input_shape[2]
        in_channel_num = input_shape[3]
        # Do object existence summary.
        object_vector = tf.nn.max_pool(input,
                                       [1, fmap_h, fmap_w, 1],
                                       [1, fmap_h, fmap_w, 1],
                                       "VALID")
        # Reshape input to remove the one dim axes.
        object_vector = tf.reshape(object_vector, [batch_size, in_channel_num])
        return object_vector

    def _setup(self):
        self.shape = [self.in_channel_num, self.out_channel_num]
        self.weights, self._loss = self._variable_with_weight_decay(
            'weights', self.shape)
        self.biases = self._get_variable(
            'biases',
            shape=[self.out_channel_num],
            initializer=tf.constant_initializer(0.0))


class RenderingMixtureLayer(SynapseLayer):
    """
    Implementation of *Ankit B. Patel et al A Probabilistic Framework for Deep
    Learning*.

    A rendering mixture layer is a *max-sum classifier* that passes the maximal
    log likelihood of all variations of a hidden class to the next layer. It is
    a generative model that first computes the posterior probability p(c | I),
    where c is the predicted class, and I is the input image (feature map),
    then uses the predicted class to reconstruct the input images. Formally,
    the process is a hard Expectation Maximization.
    """
    # Parameter init.
    # Reparameterize to weight and bias of neural network (TODO: check whether
    # it is necessary)
    # Reconstruct input
    # Compute reconstruction gradient
    pass


__all__ = [name for name, x in locals().items() if
           not inspect.ismodule(x)]



"""
PyTorch backend for akid.
"""
from __future__ import absolute_import
import os
import sys
import inspect

import numpy as np
import torch as th
from torch.autograd import Variable

from .. import computational_graph as cg
from akid.utils import glog

float32 = th.float32
uint8 = th.uint8
to_th_dtype = {
    int: th.long,
    float: th.float
}

DATA_FORMAT = "CHW"

# Maintain two hash table for looking up variables.
tensor_by_name = {}

# the number of checkpoints to save maximally
_max_checkpoint_count = 5
_checkpoint_count = 0
_checkpoint_name_queue = []

# PyTorch version
torch_version = float(th.__version__[0:3])
assert type(torch_version) is float, "Torch version number extraction failed. Version str: {}. Extracted value: {}".format(th.__version__, float(th.__version__[0:3]))

# Device to use. Only useful for torch newer than or equal to 0.4.
device = "cuda:0" if cg.use_cuda() else "cpu"


def remove_variable_contains_str(str):
    name_list = [name for name in tensor_by_name]
    for n in name_list:
        # Clean up the monitoring tensors created to save memory.
        if str in n:
            tensor_by_name.pop(n)


def get_variable(name=None, shape=None,
                 initializer=None, trainable=True,
                 shared=True):
    """
    Create or reuse variables. Compared with directly get a tensor from
    `Tensor`, it cached a name for the variable returned. It makes the variable
    available to be retrieved by name.

    `shared` is not used yet. They are legacy code from tensorflow, which may
    be useful when torch is used for distributed training.
    """
    if name:
        reuse = cg.get_variable_scope().reuse
        name = _get_name_with_scope(name)

        if reuse:
            # Return the variable if already allocated.
            if name in tensor_by_name:
                glog.debug("Reuse variable {}".format(name))
                return tensor_by_name[name]
        elif name in tensor_by_name:
            name = _append_num(name)

    # Allocate the variable.
    if not callable(initializer):
        shape = None
        if type(initializer) is th.Tensor:
            t = initializer
            t.requires_grad = True
        else:
            t = Tensor(initializer, requires_grad=trainable)
    else:
        t = Tensor(initializer(shape), requires_grad=trainable)

    if name:
        cache_tensor(t, name)

    t._is_variable = True

    glog.debug("Created new variable {}".format(name))

    return t


def get_all_variables():
    all_variables = {}
    for t in tensor_by_name:
        n = tensor_by_name[t]
        if type(n) is th.Tensor and n.is_leaf:
            all_variables[t] = n

    return all_variables


def retrieve_tensor(name):
    """
    Retrieve tensor by name.
    """
    return tensor_by_name[name]


def cache_tensor(tensor, name):
    tensor_by_name[name] = tensor

    # The reverse direction (get tensor by name) only works when it is a
    # variable. We occasionally cache numeric values as well, e.g. learning rate.

    if isinstance(tensor, cg.NamedValue):
        tensor.name = name
    # The instance below is to ensure we are dealing with an object, so
    # attributes could be added.
    elif torch_version < 0.4:
        if isinstance(tensor, Variable):
            tensor.name_ = name
    else:
        if isinstance(tensor, th.Tensor):
            tensor.name_ = name


def cache_tensor_auto_scope(tensor, name):
    """
    Cache `tensor` with `name`, whose scope information is prepended.
    """
    name = _get_name_with_scope(name)
    cache_tensor(tensor, name)


def save(path):
    """
    Save trainable variables and current step number to file. If a maximal
    number of checkpoints have been reached, the old ones will be deleted.

    Note that data concerning training are not all saved yet, e.g. the previous
    gradient that is used for momentum optimizer, which would make the
    continued training slightly different from the original one, but the guess
    is it does not matter much.
    """
    if not os.path.isdir(path):
        os.mkdir(path)
    tensor_by_name['step'] = cg.get_step()
    name = path + "/checkpoint-{}".format(cg.get_step())
    th.save(tensor_by_name, name)
    src = 'checkpoint-{}'.format(cg.get_step())
    dst = path + "/checkpoint"
    if os.path.exists(dst):
        os.remove(dst)
    os.symlink(src, dst)

    _checkpoint_name_queue.append(name)
    if len(_checkpoint_name_queue) >= _max_checkpoint_count:
        name = _checkpoint_name_queue.pop(0)
        if os.path.exists(name):
            os.remove(name)


def restore(path):
    """
    Restore variables from checkpoints saved under `path`.

    The logic works like tensorflow --- the recovery is name based. As long as
    the code that builds the computational graph does not change, the variable
    of each block will be created with the same name as the saved one, so when
    the network is being built, the variables would be created with the
    restored values.
    """
    global tensor_by_name
    tensor_by_name = th.load(path + "/checkpoint")
    cg.set_step(tensor_by_name.pop('step'))
    # Put the name back. Seems torch's save does not save additional names
    for k in tensor_by_name:
        # Remove GPU device id if using CPU
        if not cg.use_cuda():
            if is_variable(tensor_by_name[k]):
                v = tensor_by_name.pop(k)
                k = remove_gpu_device_id(k)
                if torch_version < 0.4:
                    v = Variable(v.data.cpu(), requires_grad=True) if v.is_cuda else v
                else:
                    v = v.to("cpu") if v.is_cuda else v
                tensor_by_name[k] = v
        if is_variable(tensor_by_name[k]):
            tensor_by_name[k].name_ = k


    cg.get_variable_scope().reuse_variables()


def remove_gpu_device_id(name):
    if ':' in name:
        name = name.split(':')[0]

    return name


def _get_name_with_scope(name):
    scope_name = cg.get_scope_name()
    if cg.use_cuda():
        name = '{}/{}:{}'.format(scope_name, name, int(th.cuda.current_device()))
    else:
        name = '{}/{}'.format(scope_name, name)
    return name


def remove_scope_from_name(name):
    """
    Remove scope and device information in the name.
    """
    if ':' in name:
        name, _ = name.split(':')

    name = name.split('/')[-1]

    return name


def append_suffix(name, suffix):
    """
    Append suffix to the name of a tensor, above the level of devices.
    """
    if ':' in name:
        name, device = name.split(':')
        name += '/' + suffix
        name += ':' + device
    else:
        # The name is a name of a tensor on CPU, add suffix directly.
        name += '/' + suffix
    return name


def cache_name_if_exist(func):
    def inner_func(*args, **kwargs):
        # Due to cooperative inheritance, name can only exist in kwargs.
        d = func(*args, **kwargs)
        if 'name' in kwargs and kwargs['name']:
            name = _get_name_with_scope(kwargs['name'])
            cache_tensor(d, name)
        return d
    return inner_func


def _append_num(name):
    """
    Similarly with tensorflow, we add numbers in names to distinguish variable
    with the same name.

    The format is to add an underscore and a number.

    TODO: no validation for bad naming now.
    """
    # The number needs to be added before device id.
    if cg.use_cuda():
        name, device_id = name.split(':')

    if name[-2] != '_':
        # means no number has been appended before
        name = "{}_1".format(name)
    else:
        # increase the No by one.
        parts = name.split('_')
        no = int(parts[-1])
        no += 1
        parts[-1] = str(no)
        name = ''
        for p in parts:
            name += p

    if cg.use_cuda():
        # Add back the device id.
        name += ":" + device_id

    return name


def get_name(v, with_device_id=True, no_scope=False):
    """
    Given an object, return its name if available.

    The purpose of the function is to give a unique identifier that can be used
    to identify a computed result.
    """
    if type(v) is th.Tensor:
        if hasattr(v, "name_"):
            if with_device_id:
                name = v.name_
            else:
                name = v.name_.split(':')[0]
        else:
            name = None
    elif isinstance(v, cg.NamedValue):
        name = v.name
    else:
        raise TypeError("Type {} not supported".format(type(v)))

    if no_scope and name is not None:
        name = remove_scope_from_name(name)

    return name


def standardize_data_format(data, old_format):
    """
    Stardardize data to PyTorch format, which is channel first.

    Args:
        data: Tensor or numpy array.
            The input data.
        old_format: str
            A string describe the original format. For example, if converting
            from Tensorflow, it would be 'hwio' for parameter. See
            `SUPPORT_DATA_FORMAT` and `SUPPORT_PARA_FORMAT` for supported
            strings.
    """
    if old_format not in cg.SUPPORT_PARA_FORMAT \
       and old_format not in cg.SUPPORT_DATA_FORMAT:
        raise ValueError("The data format {} is not well specified.".format(old_format))

    if old_format in cg.SUPPORT_PARA_FORMAT:
        out_format = 'oihw'
    else:
        out_format = 'nchw'

    if type(data) == np.ndarray:
        return np.einsum('{}->{}'.format(old_format, out_format), data)
    else:
        raise ValueError("Type {} is not supported.".format(type(data)))


def init():
    """
    Does nothing. Just to be compatible with tensorflow backend.
    """
    pass


def run(op, *args, **kwargs):
    """Run an operation with arguments. For compatibility with Tensorflow"""
    return op(*args, **kwargs)


def close():
    """
    The same as `init()`.
    """
    global tensor_by_name
    tensor_by_name = {}


def eval(t):
    """
    Convert torch tensor to numpy array.
    """
    if type(t) is list or type(t) is tuple:
        return [eval(i) for i in t]

    if isinstance(t, np.ndarray) or np.isscalar(t):
        return t

    if type(t) is cg.NamedTensorTuple:
        return cg.NamedTensorTuple(t.name, [eval(v) for v in t])

    if isinstance(t, cg.NamedScalar):
        return t

    # Convert to CPU anyway. May be redundant.
    t = t.cpu()

    if torch_version < 0.4:
        if type(t) is Variable:
            v = t.data.numpy()
        else:
            v = t.numpy()
    else:
        v = t.detach().numpy()

    if torch_version < 0.4:
        if len(v) == 1:
            # Return the value directly if it is not an array.
            v = v[0]

    return v


@cache_name_if_exist
def Tensor(t, requires_grad=False, name=None):
    """
    Refer to `Tensor` in Tensorflow backend for its semantics.

    We always copy data when creating new tensors given data `t`.
    """
    type_t = type(t)
    if type_t is np.ndarray:
        if t.dtype == np.int:
            t = th.from_numpy(t)
        else:
            t = th.from_numpy(t.astype(np.float32))
    elif np.isscalar(t):
        t = th.tensor(float(t))
    elif type_t is list or type_t is tuple:
        t_element = t[0]
        element_type = type(t_element)
        # Get the type of the first element in the iterable.
        while element_type is list or element_type is tuple:
            t_element = t_element[0]
            element_type = type(t_element)
        if element_type  == th.Tensor:
            t = th.tensor(t)
        else:
            t = th.tensor(t, dtype=to_th_dtype[element_type])
    elif type_t is th.Tensor:
        pass
    else:
        raise TypeError("Unknown type {} to create tensor.".format(type(t)))

    if cg.use_cuda():
        t = t.cuda()

    if requires_grad:
        if torch_version < 0.4:
            t =  Variable(t)
        else:
            t.requires_grad_(requires_grad)


    return t


@cache_name_if_exist
def is_tensor(T):
    return type(T) is th.Tensor or type(T) is Variable


def is_variable(T):
    if torch_version < 0.4:
        return type(T) is Variable
    else:
        if type(T) is not th.Tensor:
            return False
        elif hasattr(T, "_is_variable"):
            return T._is_variable
        # When the variable is loaded from file by torch.load, it does not have
        # the `_is_variable` attribute. In such a case, we check if the
        # variable is a leaf node.
        else:
            return T.is_leaf


@cache_name_if_exist
def mul(a, b, name=None):
    return a * b


@cache_name_if_exist
def mean(v, dim=None, keep_dim=False, name=None):
    if v.dtype != th.float:
        v = v.type(th.float)
    if dim is None:
        return th.mean(v)
    else:
        return th.mean(v, dim, keepdim=keep_dim)


@cache_name_if_exist
def exp(v, name=None):
    return th.exp(v)


@cache_name_if_exist
def abs(v, name=None):
    return th.abs(v)


def std(v, dim=None, keep_dim=False, name=None):
    if dim is None:
        return th.std(v)
    else:
        return th.std(v, dim, keepdim=keep_dim)


def var(v, dim=None, keep_dim=False, name=None):
    if dim is None:
        return th.var(v)
    else:
        return th.var(v, dim, keepdim=keep_dim)


@cache_name_if_exist
def max(v, name=None):
    return th.max(v)


@cache_name_if_exist
def log(v, name=None):
    return th.log(v)


@cache_name_if_exist
def linspace(start, end, space, name=None):
    return th.linspace(start, end, space)


def get_shape(t):
    if isinstance(t, cg.NamedScalar):
        return 0

    if isinstance(t, cg.NamedTensorTuple):
        return [get_shape(v) for v in t]

    if torch_version < 0.4:
        if type(t) is Variable:
            return list(t.data.shape)
        else:
            return list(t.shape)
    else:
        return list(t.shape)


def reshape(v, shape, name=None):
    return v.view(shape)


def split(v, chunk_size, dim=0, name=None):
    return v.split(chunk_size, dim=dim)


def convert_to_tensor(v):
    """
    Convert to tensor if necessary.
    """
    t = type(v)
    if t is Variable or t is th.Tensor:
        return v

    return th.tensor(v)


@cache_name_if_exist
def sqrt(v, name=None):
    return th.sqrt(v)


@cache_name_if_exist
def add(a, b, name=None):
    v = a + b

    return v

@cache_name_if_exist
def add_n(l, name=None):
    """
    Add all the tensors in the list.
    """
    acc = l[0].clone()
    for v in l[1:]:
        acc += v

    return acc


@cache_name_if_exist
def div(v, denominator, name=None):
    return th.div(v, denominator)


@cache_name_if_exist
def cat(v, name=None):
    return th.cat(v)


@cache_name_if_exist
def gather(input, dim, index, name=None):
    return th.gather(input, dim, index)


def scatter_parallel(data, devices):
    if torch_version < 0.4:
        return th.nn.parallel._functions.Scatter(devices, dim=0)(data)
    else:
        return th.nn.parallel._functions.Scatter.apply(devices, None, 0, data)


@cache_name_if_exist
def gather_parallel(data, output_device, name=None):
    if torch_version < 0.4:
        return th.nn.parallel._functions.Gather(output_device, dim=0)(*data)
    else:
        return th.nn.parallel._functions.Gather.apply(output_device, 0, *data)


@cache_name_if_exist
def expand_dims(x, axis, name=None):
    return th.unsqueeze(x, dim=axis)


@cache_name_if_exist
def transpose(x, name=None):
    return x.t()


@cache_name_if_exist
def squeeze(x, dim=None, name=None):
    if dim is not None:
        return th.squeeze(x, dim=dim)
    else:
        return th.squeeze(x)


@cache_name_if_exist
def cast(x, dtype, name=None):
    return x.type(dtype)


@cache_name_if_exist
def stack(x, name=None):
    return th.stack(x)


@cache_name_if_exist
def value(x, name=None):
    """
    Get the value of a Tensor.
    """
    return x.detach().clone()


@cache_name_if_exist
def zeros(shape, name=None):
    return th.zeros(shape, device=device)


@cache_name_if_exist
def ones(shape, name=None):
    return th.ones(shape, device=device)


@cache_name_if_exist
def eye(n, name=None):
    return th.eye(n, device=device)


@cache_name_if_exist
def range(start, end, name=None):
    return th.arange(start, end)


@cache_name_if_exist
def symeig(H):
    return th.symeig(H, eigenvectors=True)


@cache_name_if_exist
def randn(shape, name=None):
    return th.randn(*shape, device=device)


@cache_name_if_exist
def randint(low, high, size=1, name=None):
    return th.randint(low=low, high=high, size=size, device=device)


@cache_name_if_exist
def pad(x, paddings, name=None):
    # Reverse the padding para, since torch uses it backwardly.
    paddings.reverse()
    padding_flattened = []
    for i, v in enumerate(paddings):
        padding_flattened.extend(v)
    padding_flattened = tuple(padding_flattened)
    return th.nn.functional.pad(x, padding_flattened)


@cache_name_if_exist
def matmul(a, b, name=None):
    return th.matmul(a, b)


attributes = dir(sys.modules[__name__])
public_attributes = []
for a in attributes:
    if a[0] == '_' or inspect.ismodule(getattr(sys.modules[__name__], a)):
        continue
    public_attributes.append(a)
__all__ = public_attributes



"""A simple Google-style logging wrapper.

Adopted from: https://github.com/benley/python-glog.

This library attempts to greatly simplify logging in Python applications.
Nobody wants to spend hours pouring over the PEP 282 logger documentation, and
almost nobody actually needs things like loggers that can be reconfigured over
the network.  We just want to get on with writing our apps.

## Core benefits

* You and your code don't need to care about how logging works. Unless you want
  to, of course.

* No more complicated setup boilerplate!

* Your apps and scripts will all have a consistent log format, and the same
  predictable behaviours.

* It is designed to work with `akid`, so akid can centralize control of
  logging, by indicating `akid_logger` as True when calling `init`.

## Behaviours

* No default behavior, since in python logging, once a handler is added, it
  cannot be removed.

* Messages are always written to `stderr` if `glog.init()` is used without any
  arugment.

* By calling `glog.init(FILE_NAME)`, where FILE_NAME is a `str`, logs will be
  saved to that file. Target files only need to be initialized once and could
  be shared anywhere. Repeated initialization is supported, and all logs will
  be added to that file only once.

* Calling `glog.init("stderr")` or `glog.init("stdout")` will make glog log to
  standard error or standard output.

* Lines are prefixed with a google-style log prefix, of the form

      E0924 22:19:15.123456 19552 filename.py:87] Log message blah blah

  Splitting on spaces, the fields are:

    1. The first character is the log level, followed by MMDD (month, day)
    2. HH:MM:SS.microseconds
    3. Process ID
    4. basename_of_sourcefile.py:linenumber]
    5. The body of the log message.


## Example use

    import glog as log

    log.info("It works.")
    log.warn("Something not ideal")
    log.error("Something went wrong")
    log.fatal("AAAAAAAAAAAAAAA!")

If your app uses gflags, it will automatically gain a --verbosity flag. In
order for that flag to be effective, you must call log.init() after parsing
flags, like so:

    import sys
    import gflags
    import glog as log

    FLAGS = gflags.FLAGS

    def main():
      log.debug('warble garble %s', FLAGS.verbosity)

    if __name__ == '__main__':
        posargs = FLAGS(sys.argv)
        log.init()
        main(posargs[1:])

Happy logging!
"""
from __future__ import absolute_import
import sys
import logging
import time
import gflags
import types
import os

from .tools import currentframe

gflags.DEFINE_integer('verbosity', logging.INFO, 'Logging verbosity.',
                      short_name='v')
FLAGS = gflags.FLAGS
file_names = []


def format_message(record):
    try:
        record_message = '%s' % (record.msg % record.args)
    except TypeError:
        record_message = record.msg
    return record_message


class GlogFormatter(logging.Formatter):
    LEVEL_MAP = {
        logging.FATAL: 'F',
        logging.ERROR: 'E',
        logging.WARN: 'W',
        logging.INFO: 'I',
        logging.DEBUG: 'D'
    }

    def __init__(self):
        logging.Formatter.__init__(self)

    def format(self, record):
        try:
            level = GlogFormatter.LEVEL_MAP[record.levelno]
        except:
            level = '?'
        date = time.localtime(record.created)
        date_usec = (record.created - int(record.created)) * 1e6
        record_message = '%c%02d%02d %02d:%02d:%02d.%06d %s %s:%d] %s' % (
            level, date.tm_mon, date.tm_mday, date.tm_hour, date.tm_min,
            date.tm_sec, date_usec,
            record.process if record.process is not None else '?????',
            record.filename,
            record.lineno,
            format_message(record))
        record.getMessage = lambda: record_message
        return logging.Formatter.format(self, record)

logger = None


def setLevel(newlevel):
    logger.setLevel(newlevel)
    FLAGS.verbosity = newlevel
    logger.debug('Log level set to %s', newlevel)


def init(filename=None, akid_logger=False):
    """
    If `akid_logger` is True, the logger is used for logging in `akid`, in
    which case the frame will be traced back to the caller of `Block.log` or
    any `log` of its subclasses.
    """
    global logger, debug, info, warning, warn, error, exception, fatal, log

    if logger:
        return

    logger = logging.getLogger("glog")

    if akid_logger:
        def findCaller(self, *args, **kwargs):
            """
            Find the stack frame of the caller so that we can note the source
            file name, line number and function name.
            """
            f = currentframe()
            #On some versions of IronPython, currentframe() returns None if
            #IronPython isn't run with -X:Frames.
            if f is not None:
                f = f.f_back
            rv = "(unknown file)", 0, "(unknown function)"
            while hasattr(f, "f_code"):
                co = f.f_code
                filename = os.path.normcase(co.co_filename)
                # Backtrack to akid
                if "akid" not in filename:
                    f = f.f_back
                    continue
                # Backtrack to methods that are not called `log`
                if co.co_name == "log":
                    f = f.f_back
                    continue
                rv = (co.co_filename, f.f_lineno, co.co_name, None)
                break
            return rv

        logger.findCaller = types.MethodType(findCaller, logger)

    logger.propagate = False

    add(filename)
    setLevel(FLAGS.verbosity)

    debug = logger.debug
    info = logger.info
    warning = logger.warning
    warn = logger.warning
    error = logger.error
    exception = logger.exception
    fatal = logger.fatal
    log = logger.log


def add(filename):
    """
    Add file to write the log for.
    """
    if filename is None:
        if "stderr" not in file_names:
            handler = logging.StreamHandler()
            filename = "stderr"
    elif filename in file_names:
        # Do not add files that already has been added.
        return
    elif filename == "stderr":
        handler = logging.StreamHandler(sys.stderr)
    elif filename == "stdout":
        handler = logging.StreamHandler(sys.stdout)
    else:
        handler = logging.FileHandler(filename)

    file_names.append(filename)
    handler.setFormatter(GlogFormatter())
    logger.addHandler(handler)


DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
WARN = logging.WARN
ERROR = logging.ERROR
FATAL = logging.FATAL

# basicConfig = logger.basicConfig

_level_names = {
    DEBUG: 'DEBUG',
    INFO: 'INFO',
    WARN: 'WARN',
    ERROR: 'ERROR',
    FATAL: 'FATAL'
}

_level_letters = [name[0] for name in _level_names.values()]

GLOG_PREFIX_REGEX = (
    r"""
    (?x) ^
    (?P<severity>[%s])
    (?P<month>\d\d)(?P<day>\d\d)\s
    (?P<hour>\d\d):(?P<minute>\d\d):(?P<second>\d\d)
    \.(?P<microsecond>\d{6})\s+
    (?P<process_id>-?\d+)\s
    (?P<filename>[a-zA-Z<_][\w._<>-]+):(?P<line>\d+)
    \]\s
    """) % ''.join(_level_letters)
"""Regex you can use to parse glog line prefixes."""


def get_random_log_dir():
    # Naming log dir according to time if not specified.
    log_dir = "log/" + time.ctime()
    # As ':' is widely used in network protocols, replace it with '_'
    # to avoid conflict.
    log_dir = log_dir.replace(':', '_')

    return log_dir



"""
This module contains `Kid` class to play the survival game.
"""
from __future__ import absolute_import, division, print_function

import os
import time
import sys
import inspect
from tqdm import tqdm
import akid as K

from ..utils import glog as log
from . import sensors
from . import engines
from .blocks import Block
from .. import backend as A
from .events import EarlyStoppingEvent
from .common import (
    TRAIN_SUMMARY_COLLECTION,
    VALID_SUMMARY_COLLECTION,
    TRAINING_DYNAMICS_COLLECTION,
    DEFAULT_COLLECTION,
    BATCH_MONITORING_COLLECTION
)
from .eval_blocks import BatchEvalBlock
from six.moves import range


def scavenger(f):
    """
    Decorator to clean up when exception occurs.
    """
    def robust_f(self, *args, **kwargs):
        try:
            return f(self, *args, **kwargs)
        except Exception as e:
            self.teardown()
            A.reset()
            raise e
    return robust_f


class Kid(Block):
    """
    Kid is a class to assemble a `Sensor`, for supplying data, a
    `Brain`, for data processing, and a genre of `KongFu`, for algorithms or
    polices to train.

    It supports parallelism by specifying different engines.

    It has the following hooks:

        * `on_train_log_step`.
        * `on_val_log_step`
        * `on_train_begin`
        * `on_batch_begin`

    Refer to function that calls functions on hooks for detailed explanation on
    what does those hooks do. For example, to refer to method
    `on_train_log_step` for more on the hook `on_train_log_step`.

    To add a function to one of those hooks, append the function to
    `hooks.hook_name`. For example, to add a hook to `on_train_log_step`, call

    ```
    def func(kid):
        ...

    hooks.on_train_log_step.append(func)
    ```

    where `func` is the function you want it to be called. Functions added to
    hooks are supposed to take a `Kid` instance, which serves to provide
    information needed. That is also to say, no more information is available.
    """
    def __init__(self,
                 sensor_in,
                 brain_in,
                 kongfu_in,
                 engine={"name": "single"},
                 max_steps=None,
                 max_epoch=None,
                 log_dir=None,
                 log_to_file=True,
                 val_log_step=1000,
                 train_log_step=100,
                 log_by_epoch=False,
                 num_epoch_per_log=1,
                 log_by_step=True,
                 save_chk_point=True,
                 continue_from_chk_point=False,
                 inference_mode=False,
                 do_summary=True,
                 do_summary_on_val=False,
                 skip_validation=False,
                 do_batch_monitoring=False,
                 debug=False):
        """
        Assemble a sensor, a brain, and a KongFu to start the survival game.

        Args:
            sensor: Sensor
                A `Sensor` class to supply data.
            brain_in: Brain
                A `Brain` class to process data.
            kongfu_in: KongFu
                A `KongFu` class for training.
            engine: a str or a dict
                When it is a str, it should be the name of the `Engine` class
                to use, which implements parallel scheme. Available engines
                are:

                    'single', 'data_parallel'

                Default parameters of that scheme will be used.

                When it is a dict, it should be of the form:

                    {"name": "single"}
                    {"name": "data_parallel", "num_gpu": 2}

               where the `name` key indicates the parallel scheme while other
               keys are parameters of that scheme. If parameters are not
               provided, again default ones will be used.
            max_steps: int
            max_epoch: int
                You can only specify either max epoch to train or max
                steps. Not both.
            log_dir: str
                The folder to hold tensorboard event, training logs and trained
                models. If not given, first a folder named `log` will be
                created, then a folder named by current time stamp will be the
                folder to keep all the mentioned files.
            log_to_file: Boolean
                Whether to save log to file.
            val_log_step: int
                After how many steps evaluation on the validation dataset
                should be taken.
            train_log_step: int
                After how many steps training statistics should be logged.
            log_by_epoch: bool
                Whether to log at the end of each epoch.
            num_epoch_per_log: int
                Log per `epoch_num_per_log` number of epochs.
            log_by_step: bool
                Whether to log at every `val_log_step`.
            inference_mode: bool
                If the kid is set up as inference mode, stuffs related to
                training will not be built.
            do_summary: Boolean
                If False, no tensorboard summaries will be saved at all.
            do_summary_on_val: Boolean
                Whether to collect summary on the validation brain. This flag
                has no effect when `do_summary` is False. This option is for
                debugging purpose. It is useful to see how summary in
                activation of validation brain is different from training
                brain. However, if such option is used, accuracy on validation
                set will become inaccurate. This is because input from
                validation source is needed when doing summaries on validation
                brain, which would make validation data be used. Since
                validation source will reshuffle data after one epoch is
                finished, some validation may be reused and some may not be
                seen at all when doing the actual validation.
            skip_validation: bool
                If True, would not do validation during training.
            do_batch_monitoring: bool
                Whether to do summary on a specific batch throughout
                training. If you decide to switch this option on, you also need
                to manually add summary ops when building the network to the
                collection `BATCH_MONITORING_COLLECTION`, otherwise, nothing
                would be done.
            Other args are self-evident.
        """
        self.sensor = sensor_in
        self.brain = brain_in
        self.kongfu = kongfu_in
        self.engine_para = engine
        self.debug = debug

        # Set up logging facilities.
        if log_dir is None:
            self.log_dir = log.get_random_log_dir()
        else:
            self.log_dir = os.path.normpath(log_dir)
        self.log_filepath = self.log_dir + "/training.log"
        self.model_dir = self.log_dir + "/model"
        self.log_to_file = log_to_file

        self.max_steps = max_steps
        self.max_epoch = max_epoch
        assert self.max_steps is None or self.max_epoch is None,\
            "Only one of `max_steps` and `max_epoch` could be used."
        assert self.max_steps is not None or self.max_epoch is not None,\
            "At least one `max_steps` and `max_epoch` is needed."

        self.train_log_step = train_log_step
        self.val_log_step = val_log_step
        self.log_by_epoch = log_by_epoch
        self.num_epoch_per_log = num_epoch_per_log
        self.log_by_step = log_by_step
        self.inference_mode = inference_mode
        self.do_summary_on_val = do_summary_on_val
        self.do_summary = do_summary
        self.save_chk_point = save_chk_point
        self.continue_from_chk_point = continue_from_chk_point
        self.skip_validation = skip_validation
        self.do_batch_monitoring = do_batch_monitoring

        # Log training dynamics
        self.loss_data_train = []
        self.eval_data_train = []

        self.loss_data_val = []
        self.eval_data_val = []

        self.initialized = False

        # Set up hooks.
        class hooks(object):
            def __init__(self):
                self.on_training_log_step = []
                self.on_val_log_step = []
                self.on_train_begin = []
                self.on_batch_begin = []
                self.on_epoch_end = []
                self.add_default_hooks()

            def add_default_hooks(self):
                from .callbacks import on_train_log_step
                self.on_training_log_step.append(on_train_log_step)

                from .callbacks import on_val_log_step
                self.on_val_log_step.append(on_val_log_step)

                from .callbacks import on_train_begin
                self.on_train_begin.append(on_train_begin)

                from .callbacks import on_batch_begin
                self.on_batch_begin.append(on_batch_begin)

                from .callbacks import on_epoch_end
                self.on_epoch_end.append(on_epoch_end)

        self.hooks = hooks()

        # Class members whose value depends on the state of the class.
        self.loss = None
        self.evals = None
        self.best_val_evals = None

    @scavenger
    def validate(self, mode="val"):
        """Evaluating on validation set.

        Args:
            mode: str
                If "val", validate on validation set; if "test", validate on
                test set.

        Return:
            loss: float
                The validation loss. Note that though only loss is returned,
                `Kid`'s reference to evaluation metrics and loss are both
                updated.
        """
        self.init()

        A.check_mode(mode)

        if mode == A.Mode.VAL:
            self.log('Validation Data Eval:')
        elif mode == A.Mode.TEST:
            self.log('Test Data Eval:')
        elif mode == A.Mode.TRAIN:
            self.log('Train Data Eval:')
        else:
            assert False, "Program should not reach here."

        if not self.sensor.mode == mode:
            self.sensor.set_mode(mode)

        # We want to traverse the dataset, so reset to reset the sampler.
        self.sensor.reset()

        # Run one epoch of eval.
        self.log("A epoch of {} set contains {} batches. Batch size {}.".format(mode, self.sensor.num_batches_per_epoch, self.sensor.batch_size))
        eval_blocks = [BatchEvalBlock()
                       if K.get_eval_block(A.get_name(v, no_scope=True)) is None
                       else K.get_eval_block(A.get_name(v, no_scope=True))()
                       for v in self.engine.eval(get_val=True)]
        if self.engine.verbose_eval(get_val=True) is not None:
            verbose_eval_blocks = [BatchEvalBlock()
                                   if K.get_eval_block(A.get_name(v, no_scope=True)) is None
                                   else K.get_eval_block(A.get_name(v, no_scope=True))()
                                   for v in self.engine.verbose_eval(get_val=True)]
        else:
            verbose_eval_blocks = None
        loss_block = BatchEvalBlock()
        steps_per_epoch = self.sensor.num_batches_per_epoch

        for step in tqdm(list(range(1, steps_per_epoch+1))):
            if verbose_eval_blocks is None:
                loss, evals = self.run_step(update=False, val=True)
            else:
                loss, evals, verbose_evals = self.run_step(update=False, val=True)

            loss_block.add(loss)
            for i, v in enumerate(evals):
                eval_blocks[i].add(v)
            if verbose_eval_blocks is not None:
                for i, v in enumerate(verbose_evals):
                    verbose_eval_blocks[i].add(v)

        # Note that the EvalBlocks become numeric data now.
        loss_avg = loss_block.data
        for i, v in enumerate(eval_blocks):
            eval_blocks[i] = v.data
        if verbose_eval_blocks is not None:
            for i, v in enumerate(verbose_eval_blocks):
                verbose_eval_blocks[i] = v.data

        self.loss = loss_avg
        self.evals = eval_blocks
        self.verbose_evals = verbose_eval_blocks

        return loss_avg, eval_blocks

    def reset(self):
        self.sensor.reset()
        self.sensor.setup()

    @scavenger
    def setup(self):
        """
        Set up logging and the computation graph.
        """
        self._setup_log()

        if A.backend() == A.TORCH and self.continue_from_chk_point:
            log.info("Recovering net from checkpoint %s." % self.model_dir)
            self.restore()

        if self.sensor.do_summary is None:
            self.sensor.set_do_summary_flag(self.do_summary)
        if self.sensor.do_summary_on_val is None:
            self.sensor.set_do_summary_on_val_flag(self.do_summary_on_val)
        if self.brain.do_summary is None:
            self.brain.set_do_summary_flag(self.do_summary)
        if self.brain.do_summary_on_val is None:
            self.brain.set_do_summary_on_val_flag(self.do_summary_on_val)

        self.sensor.setup()
        self.engine = engines.get(brain=self.brain, kongfu=self.kongfu, **self.engine_para)
        self.engine.setup()

        # Do forward once to build ops.

        self.step(update=False)
        # Build validation ops
        self.sensor.set_mode("val")
        self.sensor.setup()
        self.step(update=False, val=True)

    def teardown(self):
        self.sensor.teardown()

    def restore(self):
        A.restore(self.model_dir)
        if A.backend() ==  A.TORCH and not self.inference_mode:
            self.kongfu.set_lr(A.retrieve_tensor('lr'))

    def batch_monitoring(self, data):
        """
        A method that creates summary ops to visualize a specific batch
        throughout training to monitor the progress of training.

        Args:
            data:
                The specific data batch to monitor.
        """
        # Propagate the sample through network.
        # NOTE: the following code is PyTorch specific.
        self.brain.switch_batch_monitoring_mode()
        self.step(update=False, val=False, data=data)

        # Create summary ops if not already, and run them.
        if not hasattr(self, "batch_monitoring_summary_ops"):
            self.batch_monitoring_summary_ops = A.summary.get_collection(BATCH_MONITORING_COLLECTION)
        A.summary.run_summary_op(self.batch_monitoring_summary_ops)

        # Clean up the monitoring tensors created to save memory.
        # A.remove_variable_contains_str(self.brain.name)

        self.brain.switch_batch_monitoring_mode()

        if not self.brain.done_first_batch_monitoring_pass:
            self.brain.set_flag("done_first_batch_monitoring_pass", True)


    @scavenger
    def practice(self, return_eval=False):
        """
        Improve the performance of the kid's brain by practicing, aka
        applying back propagation to train neural network.

        Args:
            return_eval: bool
                Return evaluation metric after trainning is finished.
        Return:
            loss, [eval]: float, float
                Final validation loss and optional evaluation metric.
        """
        self.log("Begin training brain: " + self.brain.name)
        self.sensor.set_mode("train")
        self.sensor.setup()
        self.log("A epoch of training set contains {} batches".format(self.sensor.num_batches_per_epoch))
        self.epoch = A.get_step() // self.sensor.num_batches_per_epoch

        self.init()
        self.on_train_begin()
        # Reset sensor, so to prevent any code from using some training batches
        # in advance, and making them missing from initial training.
        self.sensor.reset()

        val_loss, val_evals = None, None
        while A.get_step() < self.max_steps:
            try:
                val_loss, val_evals = self.step_with_logistics()
            except EarlyStoppingEvent as e:
                val_loss, val_evals = e.val_loss, e.val_evals
                break

        if return_eval:
            return val_loss, val_evals
        else:
            return val_loss

    @scavenger
    def predict(self, data):
        """
        Given a datum, do inference with a NN. Note that we do not have format
        requirement on `data`, since it should be taken care when building the
        NN. If input validation is needed, it should be done in the layer that
        calls this API.
        """
        self.init()
        self.engine.set_val(True)
        self.engine.forward(data, val=True)

        evals = [e for e in self.engine.eval(get_val=True)]
        if self.engine.verbose_eval(get_val=True) is not None:
            verbose_evals = [e for e in self.engine.verbose_eval(get_val=True)]
        else:
            verbose_evals = None

        return evals, verbose_evals

    def step_with_logistics(self):
        val_loss = None
        val_evals = None

        start_time = time.time()

        self.on_batch_begin()
        self.loss, self.evals = self.run_step()

        if self.debug:
            print("Loss: {}; Eval: {}".format(self.loss, self.evals))

        self.step_time = time.time() - start_time

        A.step()

        if A.get_step() % self.train_log_step == 0:
            # Since during logging, loss and evals are passed in as attributes
            # of Kid, training log is put ahead of validation log to prevent
            # validation loss and evals from overriding the ones of training,
            # when the validation logging step and training logging step
            # coincide.
            self.on_train_log_step()

        if self.log_by_step and \
            (A.get_step() % self.val_log_step == 0 or\
            A.get_step() == self.max_steps):
            if self.save_chk_point:
                self.save_to_ckpt()
            if not self.skip_validation:
                self.loss, self.evals = self.validate()
            val_loss, val_evals = self.loss, self.evals
            self.on_val_log_step()
            self.sensor.set_mode("train")
            self.sensor.setup()

        if A.get_step() % self.sensor.num_batches_per_epoch is 0:
            self.epoch += 1
            self.on_epoch_end()
            if self.log_by_epoch \
               and A.get_step() % (self.sensor.num_batches_per_epoch * self.num_epoch_per_log) == 0:
                if self.save_chk_point:
                    self.save_to_ckpt()
                if not self.skip_validation:
                    self.loss, self.evals = self.validate()
                self.on_val_log_step()
                val_loss, val_evals = self.loss, self.evals
                self.sensor.set_mode("train")
                self.sensor.setup()

        return val_loss, val_evals

    def tf_step(self, update=True, val=False):
        self.sensor.set_val(val)
        self.feed_dict = self.get_feed_dict(val)
        fetch = {"loss": self.engine.loss(val)}
        fetch["eval"] = self.engine.eval(val)
        if val:
            # TODO: may be broken. could dict element be list?
            fetch["verbose_eval"] = self.engine.verbose_eval(val)
        if update:
            fetch["train_op"] = self.engine.train_op
        result = A.run(fetch, feed_dict=self.feed_dict)

        return result["loss"], result["eval"], result["verbose_eval"] if val else None

    def run_step(self, update=True, val=False, data=None):
        """
        Computation wise, execute the computational graph for a step.
        """
        if A.backend() == A.TF:
            return self.tf_step(update, val)
        elif A.backend() == A.TORCH:
            loss, evals, verbose_evals = self.step(update, val, data=data)
            loss = A.eval(loss)
            evals = A.eval(evals)
            if verbose_evals is not None and val:
                verbose_evals = A.eval(verbose_evals)
                return loss, evals, verbose_evals
            else:
                return loss, evals

    def step(self, update=True, val=False, data=None):
        """
        Computational graph wise, how the tensor should be run in a step.

        If `data` is None, the caller is supposed to feed in data. In such a
        case, the control on the sensor is left to the caller; that is, kid
        would not modify it after setup.
        """
        if data is None:
            system_in = self.sensor.forward()
        else:
            system_in = data

        self.engine.set_val(val)
        self.engine.forward(system_in, val)
        if update:
            self.engine.update()

        loss = self.engine.loss(val)
        evals = [e for e in self.engine.eval(val)]
        if val and self.engine.verbose_eval(val) is not None:
            verbose_evals = [e for e in self.engine.verbose_eval(val)]
        else:
            verbose_evals = None

        return loss, evals, verbose_evals

    def _setup_log(self):
        if not os.path.exists(self.log_dir):
            os.makedirs(self.log_dir)
        if not os.path.exists(self.model_dir):
            os.makedirs(self.model_dir)
        A.init_log()
        if self.log_to_file:
            log.add(self.log_filepath)
        if self.debug:
            log.setLevel(log.DEBUG)
        else:
            log.setLevel(log.INFO)
        self.log("Logs will be save to: {}".format(self.log_dir))

    def init(self):
        """
        Initialize the backend if having not.
        """
        if not self.initialized:
            if A.backend() == A.TF:
                # The checkpoint recovery has to be here, given that it needs to be
                # after the computational graph being set up.
                A.init(self.continue_from_chk_point, self.model_dir)
                self.initialized = True
            elif A.backend() == A.TORCH:
                A.init() # This actually does nothing.
                self.initialized = True
            else:
                raise ValueError("Backend {} not supported".format(A.backend()))

    def setup_summary(self):
        if self.do_summary:
            # SummaryWriter to output summaries and the Graph.
            A.summary.init(self.log_dir)

            summary_ops = A.summary.get_collection(TRAIN_SUMMARY_COLLECTION)
            summary_ops.extend(A.summary.get_collection(
                TRAINING_DYNAMICS_COLLECTION))
            summary_ops.extend(A.summary.get_collection(
                DEFAULT_COLLECTION))
            if self.do_summary_on_val:
                val_summary_ops = A.summary.get_collection(
                    VALID_SUMMARY_COLLECTION)
                summary_ops.extend(val_summary_ops)
            self.summary_op = A.summary.merge(summary_ops)
            A.summary.add_graph()

    def save_to_ckpt(self):
        # For torch background, the learning rate will be used during setup
        # phase, so it should be saved.
        if A.backend() == A.TORCH:
            A.cache_tensor(self.kongfu.get_lr(), 'lr')
        A.save(self.model_dir)
        self.log("Checkpoint at step {} saved to folder:"
                 " {}".format(A.get_step(), self.model_dir))

    def get_feed_dict(self, val=False):
        feed_dict = None
        if type(self.sensor) is sensors.FeedSensor:
            # Placeholder of `FeedSensor` should be filled.
            feed_dict = self.sensor.fill_feed_dict()
            # If val is True, we have gotten val data already.
            if not val and self.do_summary and self.do_summary_on_val:
                # Validation data is also needed, so add them in.
                self.sensor.set_val(True)
                val_feed_dict = self.sensor.fill_feed_dict()
                self.sensor.set_val(False)
                feed_dict.update(val_feed_dict)

        lr_dict = self.kongfu.get_feed_dict()

        if feed_dict is not None:
            feed_dict.update(lr_dict)
        else:
            feed_dict = lr_dict

        return feed_dict

    def on_train_log_step(self):
        """
        Call hooks at the time when the kid should do logging for training.
        """
        for func in self.hooks.on_training_log_step:
            func(self)

    def on_val_log_step(self):
        """
        Call hooks at the time when the kid should do logging for validation.
        """
        # Update the best validation evaluation results.
        if self.best_val_evals:
            for i, e in enumerate(self.evals):
                if self.best_val_evals[i] < e:
                    self.best_val_evals[i] = e
        else:
            self.best_val_evals = list(self.evals)

        for func in self.hooks.on_val_log_step:
            func(self)

    def on_train_begin(self):
        if self.max_epoch:
            # Convert the max epoch number to max steps.
            self.max_steps \
                = self.sensor.num_batches_per_epoch * self.max_epoch

        for func in self.hooks.on_batch_begin:
            func(self)
        for func in self.hooks.on_train_begin:
            func(self)

    def on_batch_begin(self):
        """
        NOTE: all functions attached to this hook will be called in the hook
        `on_train_begin` as well, with a higher priority.
        """
        for func in self.hooks.on_batch_begin:
            func(self)

    def on_epoch_end(self):
        for func in self.hooks.on_epoch_end:
            func(self)

__all__ = [name for name, x in locals().items() if
           not inspect.ismodule(x) and not inspect.isabstract(x)]



from __future__ import absolute_import
from akid.sugar.block_templates import cnn_block
from akid import AKID_DATA_PATH
from akid import GraphBrain, MNISTFeedSource, FeedSensor, Kid, MomentumKongFu
from akid.layers import SoftmaxWithLossLayer
from akid import LearningRateScheme


def setup(bn=None, activation_before_pooling=False):
    brain = GraphBrain(name="sugar_mnist")

    brain.attach(cnn_block(ksize=[5, 5],
                           initial_bias_value=0.,
                           init_para={"name": "truncated_normal",
                                      "stddev": 0.1},
                           wd={"type": "l2", "scale": 5e-4},
                           in_channel_num=1,
                           out_channel_num=32,
                           pool_size=[2, 2],
                           pool_stride=[2, 2],
                           activation={"type": "relu"},
                           activation_before_pooling=activation_before_pooling,
                           bn=bn))

    brain.attach(cnn_block(ksize=[5, 5],
                           initial_bias_value=0.,
                           init_para={"name": "truncated_normal",
                                      "stddev": 0.1},
                           wd={"type": "l2", "scale": 5e-4},
                           in_channel_num=32,
                           out_channel_num=64,
                           pool_size=[5, 5],
                           pool_stride=[2, 2],
                           activation={"type": "relu"},
                           activation_before_pooling=activation_before_pooling,
                           bn=bn))

    brain.attach(cnn_block(ksize=None,
                           initial_bias_value=0.1,
                           init_para={"name": "truncated_normal",
                                      "stddev": 0.1},
                           wd={"type": "l2", "scale": 5e-4},
                           in_channel_num=3136,
                           out_channel_num=512,
                           activation={"type": "relu"},
                           bn=bn,
                           keep_prob=0.5))

    brain.attach(cnn_block(ksize=None,
                           initial_bias_value=0.1,
                           init_para={"name": "truncated_normal",
                                      "stddev": 0.1},
                           wd={"type": "l2", "scale": 5e-4},
                           in_channel_num=512,
                           out_channel_num=10,
                           bn=bn,
                           activation=None))

    brain.attach(SoftmaxWithLossLayer(
        class_num=10,
        inputs=[{"name": "ip4", "idxs": [0]},
                {"name": "system_in", "idxs": [1]}],
        name="loss"))

    source = MNISTFeedSource(name="MNIST",
                             url='http://yann.lecun.com/exdb/mnist/',
                             work_dir=AKID_DATA_PATH + '/mnist',
                             num_train=50000,
                             num_val=5000,
                             center=True,
                             scale=True)

    sensor = FeedSensor(name='data',
                        source_in=source,
                        batch_size=64,
                        val_batch_size=100)
    kid = Kid(
        sensor,
        brain,
        MomentumKongFu(
            lr_scheme={
                "name": LearningRateScheme.exp_decay,
                "base_lr": 0.01,
                "decay_rate": 0.95,
                "num_batches_per_epoch": 468,
                "decay_epoch_num": 1},
            momentum=0.9),
        max_steps=4000)
    kid.setup()

    return kid


if __name__ == "__main__":
    kid = setup()
    kid.practice()



import sys
import os
import pdb
import functools
import traceback


def debug_on(*exceptions, on=True):
    if not exceptions:
        exceptions = (AssertionError, )
    def decorator(f):
        if not on:
            return f

        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except exceptions:
                info = sys.exc_info()
                traceback.print_exception(*info)
                pdb.post_mortem(info[2])
        return wrapper
    return decorator



"""
This module contains `Joker`s, the classes that make benign jokes to data, aka
doing data augmentations.
"""
from __future__ import absolute_import, division, print_function

import abc
import sys
import inspect

import tensorflow as tf

from .blocks import ShadowableBlock
from .systems import SequentialSystem
import akid


class Joker(ShadowableBlock):
    """
    A top level abstract class to do data augmentations. Since it is also part
    of a data processing process, it is a sub-class of `ShadowableBlock`.

    A Joker normally accepts one input -- so its `_forward` only takes one input,
    and gives out one output -- so the output is revealed via property `data`.
    """
    def __init__(self, do_summary=False, **kwargs):
        """
        Args:
            do_summary: Boolean
                Do not to do summary on output by default, since it would add
                too much relatively repeated summary since the statistics of
                data won't change much.
        """
        # do_summary is argument of the super class. We change its default
        # value, so we put the changed value in.
        kwargs["do_summary"] = do_summary
        super(Joker, self).__init__(**kwargs)

    @abc.abstractmethod
    def _forward(self, data_in):
        raise NotImplementedError("Each `Joker` should implement `_forward` to"
                                  " do actual data augmentation!")
        sys.exit()

    @property
    def data(self):
        """
        All sub-class `Joker` should save the augmented data to `_data`.
        """
        return self._data


class JokerSystem(SequentialSystem):
    """
    A system consists of linearly linked jokers to do data augmentation.
    """
    def attach(self, joker):
        assert issubclass(type(joker), Joker),\
            "A `JokerSystem` should only contain `Joker`s."
        super(JokerSystem, self).attach(joker)


class CropJoker(Joker):
    def __init__(self,
                 height=None, width=None,
                 center=False, central_fraction=None,
                 **kwargs):
        """
        A `Joker` that randomly crop a region of `height` and `width` from
        an input image.

        Args:
            center: a Boolean. Whether to resize an image to a target width
                and height by either centrally cropping the image or padding it
                evenly with zeros, or randomly crop the input image to get an
                image with desired size.
            central_fraction: float
                If existing, crop the image by fraction instead of using height
                and width passed in.
        """
        super(CropJoker, self).__init__(**kwargs)
        self.height = height
        self.width = width
        self.center = center
        self.central_fraction = central_fraction

    def _forward(self, data_in):
        if self.center:
            self.log("Center crop images.")
            if self.central_fraction:
                self._data = tf.image.central_crop(
                    data_in, central_fraction=self.central_fraction)
            else:
                self._data = tf.image.resize_image_with_crop_or_pad(
                    data_in, self.height, self.width)
        else:
            self.log("Randomly crop images.")
            shape = data_in.get_shape().as_list()
            assert self.width and self.height,\
                "crop height and width should not be None."
            self._data = tf.random_crop(data_in,
                                        [self.height, self.width, shape[-1]])

        return self._data


class FlipJoker(Joker):
    def __init__(self, flip_left_right=True, **kwargs):
        """
        A `Joker` that randomly flips input images.

        Args:
            flip_left_right: Boolean
                If True, do randomly horizontal flipping, otherwise, do
                vertical flipping.
        """
        super(FlipJoker, self).__init__(**kwargs)
        self.flip_left_right = flip_left_right

    def _forward(self, data_in):
        if self.flip_left_right:
            self.log("Randomly flip image left right.")
            self._data = tf.image.random_flip_left_right(data_in)
        else:
            self.log("Randomly flip image up down.")
            self._data = tf.image.random_flip_up_down(data_in)

        return self._data


class LightJoker(Joker):
    def __init__(self, contrast=True, brightness=True, **kwargs):
        """
        A `Joker` that randomly adjust contrast and brightness of input images.

        Args:
            contrast: Boolean
                If True, randomly adjust contrast.
            brightness: Boolean
                If True, randomly adjust brightness.
        """
        super(LightJoker, self).__init__(**kwargs)
        self.contrast = contrast
        self.brightness = brightness

    def _forward(self, data_in):
        data = data_in
        # TODO(Shuai): The parameters should not be hard coded.
        if self.contrast:
            self.log("Randomly change contrast.")
            data = tf.image.random_contrast(data, lower=0.2, upper=1.8)
        if self.brightness:
            self.log("Randomly change brightness.")
            data = tf.image.random_brightness(data, max_delta=63)

        self._data = data

        return self._data


class WhitenJoker(Joker):
    """
    Per image whitening joke.
    """
    def _forward(self, data_in):
        self._data = tf.image.per_image_standardization(data_in)
        return self._data


class RescaleJoker(Joker):
    """
    Rescale images to $[0, 1]$.
    """
    def _forward(self, data_in):
        self._data = akid.image.rescale_image(data_in)
        return self._data


class ResizeJoker(Joker):
    def __init__(self,
                 height,
                 width,
                 resize_method=0,
                 **kwargs):
        """
        A `Joker` that resizes an image to [height, width].

        Args:
            resize_method: int
                the resize method to use, see doc of `tf.image.resize_images`
                for details. Briefly, available resize methods are (retrieved
                from tensorflow version 0.11):
                    BILINEAR = 0
                    NEAREST_NEIGHBOR = 1
                    BICUBIC = 2
                    AREA = 3
        """
        super(ResizeJoker, self).__init__(**kwargs)
        self.height = height
        self.width = width
        self.resize_method = resize_method

    def _forward(self, data_in):
        shape = data_in.get_shape().as_list()
        if len(shape) == 3:
            data = tf.expand_dims(data_in, 0)
        resized_image = tf.image.resize_images(data,
                                               [self.height, self.width],
                                               method=self.resize_method)
        resized_image = tf.squeeze(resized_image)
        self._data = resized_image

        return self._data


__all__ = [name for name, x in locals().items() if
           not inspect.ismodule(x) and not inspect.isabstract(x)]



"""
An extension of `multiprocessing`'s `Semaphore` to provide batch acquire of
semaphore.
"""
from __future__ import print_function

from __future__ import absolute_import
from multiprocessing import Semaphore as _Semaphore
from multiprocessing import Lock
from six.moves import range


class Semaphore(object):
    def __init__(self, count):
        self.semaphore = _Semaphore(count)
        self.in_lock = Lock()
        self.out_lock = Lock()

    def acquire(self, count):
        """
        Args:
            count: int
                The number of semaphores to acquire.
        """
        with self.in_lock:
            for i in range(0, count):
                self.semaphore.acquire()
                print ("Acquired semaphore.")

    def release(self, count):
        """
        Args:
            count: int
                The number of semaphores to acquire.
        """
        with self.out_lock:
            for i in range(0, count):
                self.semaphore.release()
                print ("Released semaphore.")
